#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import json

from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QLabel, QVBoxLayout, QWidget, QCheckBox, QHBoxLayout, QScrollArea, \
    QSizePolicy

from pygpt_net.ui.widget.element.group import QVLine, QHLine
from pygpt_net.ui.widget.element.labels import HelpLabel, UrlLabel, IconLabel
from pygpt_net.ui.widget.option.combo import OptionCombo
from pygpt_net.ui.widget.option.input import OptionInput
from pygpt_net.utils import trans


class WebTab:
    def __init__(self, window=None):
        """
        Tab: Web

        :param window: Window instance
        """
        self.window = window
        self.params_scroll = None
        self.params_widget = None

    def setup(self):
        """Setup tab widget"""
        # get loaders list
        loaders = self.window.controller.config.placeholder.apply_by_id("llama_index_loaders_web")
        loaders_list = []
        for loader in loaders:
            k = list(loader.keys())[0]
            key = k.replace("web_", "")
            value = loader[k]
            loaders_list.append({
                key: value,
            })

        self.window.ui.nodes["tool.indexer.web.loader"] = OptionCombo(
            self.window,
            "tool.indexer",
            "web.loader",
            {
                "label": trans("tool.indexer.tab.web.loader"),
                "keys": loaders_list,
                "value": "webpage",
            }
        )

        self.window.ui.nodes["tool.indexer.web.loader"].layout.setContentsMargins(0, 0, 0, 0)
        self.window.ui.nodes["tool.indexer.web.loader.label"] = HelpLabel(trans("tool.indexer.tab.web.loader"))
        self.window.ui.add_hook("update.tool.indexer.web.loader", self.hook_loader_change)

        self.window.ui.nodes["tool.indexer.web.options.label"] = HelpLabel(trans("tool.indexer.tab.web.source"))

        config_label = HelpLabel(trans("tool.indexer.tab.web.cfg"))
        config_label.setWordWrap(False)

        config_label_layout = QHBoxLayout()
        config_label_layout.addWidget(IconLabel(QIcon(":/icons/settings_filled.svg")))
        config_label_layout.addWidget(config_label)
        config_label_layout.setAlignment(Qt.AlignLeft)

        self.window.ui.nodes["tool.indexer.web.config.label"] = QWidget()
        self.window.ui.nodes["tool.indexer.web.config.label"].setLayout(config_label_layout)

        self.window.ui.nodes["tool.indexer.web.config.help"] = UrlLabel(
            trans("tool.indexer.tab.web.help"),
            "https://pygpt.readthedocs.io/en/latest/configuration.html#data-loaders")

        # --------------------------------------------------

        params_layout = QVBoxLayout()
        params_layout.setContentsMargins(0, 0, 0, 0)

        self.params_scroll = QScrollArea()
        self.params_scroll.setWidgetResizable(True)
        self.window.ui.nodes["tool.indexer.web.loader.option"] = {}
        self.window.ui.nodes["tool.indexer.web.loader.option_group"] = {}
        self.window.ui.nodes["tool.indexer.web.loader.config"] = {}
        self.window.ui.nodes["tool.indexer.web.loader.config_group"] = {}

        # params
        params_layout.addWidget(self.window.ui.nodes["tool.indexer.web.options.label"])
        inputs, groups = self.window.core.idx.ui.loaders.setup_loader_options()
        for loader in inputs:
            for k in inputs[loader]:
                self.window.ui.nodes["tool.indexer.web.loader.option." + loader + "." + k] = inputs[loader][k]
        for loader in groups:
            self.window.ui.nodes["tool.indexer.web.loader.option_group"][loader] = groups[loader]
            params_layout.addWidget(self.window.ui.nodes["tool.indexer.web.loader.option_group"][loader])
            self.window.ui.nodes["tool.indexer.web.loader.option_group"][loader].hide()  # hide on start

        # separator
        params_layout.addWidget(QHLine())

        # config
        params_layout.addWidget(self.window.ui.nodes["tool.indexer.web.config.label"])
        inputs, groups = self.window.core.idx.ui.loaders.setup_loader_config()
        for loader in inputs:
            for k in inputs[loader]:
                self.window.ui.nodes["tool.indexer.web.loader.config." + loader + "." + k] = inputs[loader][k]
        for loader in groups:
            self.window.ui.nodes["tool.indexer.web.loader.config_group"][loader] = groups[loader]
            params_layout.addWidget(self.window.ui.nodes["tool.indexer.web.loader.config_group"][loader])
            self.window.ui.nodes["tool.indexer.web.loader.config_group"][loader].hide()  # hide on start
        params_layout.addWidget(self.window.ui.nodes["tool.indexer.web.config.help"], alignment=Qt.AlignCenter)

        # stretch
        params_layout.addStretch(1)

        self.params_widget = QWidget()
        self.params_widget.setLayout(params_layout)
        self.params_scroll.setWidget(self.params_widget)

        # resize
        self.params_scroll.setWidgetResizable(True)
        self.params_scroll.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)

        # --------------------------------------------------

        self.window.ui.nodes["tool.indexer.web.options.replace"] = QCheckBox(
            trans("tool.indexer.option.replace"))
        self.window.ui.nodes["tool.indexer.web.options.replace"].setChecked(True)

        self.window.ui.nodes["tool.indexer.provider"] = HelpLabel(self.window.core.config.get("llama.idx.storage"))

        loader_layout = QVBoxLayout()
        loader_layout.addWidget(self.window.ui.nodes["tool.indexer.web.loader.label"])
        loader_layout.addWidget(self.window.ui.nodes["tool.indexer.web.loader"])
        loader_layout.setContentsMargins(0, 10, 0, 0)

        options_layout = QVBoxLayout()
        options_layout.addWidget(self.window.ui.nodes["tool.indexer.web.options.replace"])

        self.window.ui.nodes["tool.indexer.web.header.tip"] = HelpLabel(trans("tool.indexer.tab.web.tip"))

        # layout
        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.nodes["tool.indexer.web.header.tip"])
        layout.addLayout(loader_layout)
        layout.addLayout(options_layout)
        layout.addWidget(self.params_scroll)

        # defaults
        self.window.ui.nodes["tool.indexer.web.loader"].set_value("webpage")

        widget = QWidget()
        widget.setLayout(layout)

        return widget

    def hook_loader_change(self, key, value, caller, *args, **kwargs):
        """
        Hook: on loader change

        :param key: Option key
        :param value: Option value
        :param caller: Caller
        :param args: Args
        :param kwargs: Kwargs
        """
        # hide/show options
        for loader in self.window.ui.nodes["tool.indexer.web.loader.option_group"]:
            self.window.ui.nodes["tool.indexer.web.loader.option_group"][loader].hide()
        if value in self.window.ui.nodes["tool.indexer.web.loader.option_group"]:
            self.window.ui.nodes["tool.indexer.web.loader.option_group"][value].show()

        # show/hide label if options are available
        self.window.ui.nodes["tool.indexer.web.options.label"].hide()
        if value in self.window.ui.nodes["tool.indexer.web.loader.option_group"]:
            self.window.ui.nodes["tool.indexer.web.options.label"].show()

        # hide/show config
        for loader in self.window.ui.nodes["tool.indexer.web.loader.config_group"]:
            self.window.ui.nodes["tool.indexer.web.loader.config_group"][loader].hide()
        if value in self.window.ui.nodes["tool.indexer.web.loader.config_group"]:
            self.window.ui.nodes["tool.indexer.web.loader.config_group"][value].show()

        # show/hide label if config are available
        self.window.ui.nodes["tool.indexer.web.config.label"].hide()
        self.window.ui.nodes["tool.indexer.web.config.help"].hide()
        if value in self.window.ui.nodes["tool.indexer.web.loader.config_group"]:
            self.window.ui.nodes["tool.indexer.web.config.label"].show()
            self.window.ui.nodes["tool.indexer.web.config.help"].show()

        self.params_widget.adjustSize()
        self.params_scroll.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import datetime
import json
from typing import Dict, Any, List

from PySide6.QtCore import Slot
from PySide6.QtGui import QAction, QIcon, QTextCursor
from sqlalchemy import text

from pygpt_net.tools.base import BaseTool
from pygpt_net.utils import trans

from .ui.dialogs import DialogBuilder


class IndexerTool(BaseTool):
    def __init__(self, *args, **kwargs):
        """
        Indexer tool

        :param window: Window instance
        """
        super(IndexerTool, self).__init__(*args, **kwargs)
        self.id = "indexer"  # unique tool id
        self.dialog = None
        self.opened = False
        self.current_idx = "base"

    def attach(self, window):
        """
        Attach window to tool

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Setup tool"""
        self.window.idx_logger_message.connect(self.handle_log)  # connect logger

    def update(self):
        """Update"""
        pass

    def open(self):
        """Open dialog window"""
        self.window.ui.dialogs.open('tool.indexer', width=800, height=600)
        self.opened = True
        self.refresh()
        self.update()

    def close(self):
        """Close dialog window"""
        self.window.ui.dialogs.close('tool.indexer')
        self.opened = False

    def toggle(self):
        """Toggle dialog window"""
        if self.opened:
            self.close()
        else:
            self.open()

    def open_settings(self):
        """Open settings dialog"""
        self.window.controller.settings.open_section("llama-index")

    def show_hide(self, show: bool = True):
        """
        Show/hide dialog window

        :param show: show/hide
        """
        if show:
            self.open()
        else:
            self.close()

    def on_close(self):
        """On dialog close"""
        self.opened = False

    def on_exit(self):
        """On app exit"""
        pass

    def on_tab_changed(self, index):
        """
        On tab changed

        :param index: tab index
        """
        if index > 1:
            self.window.ui.nodes["tool.indexer.btn.idx"].hide()
        else:
            self.window.ui.nodes["tool.indexer.btn.idx"].show()
        self.refresh()

    def set_current_idx(self, idx, check: bool = True):
        """
        Set current index

        :param idx: index id
        :param check: check if index is valid
        """
        self.current_idx = idx
        self.refresh(check)

    def check_current_idx(self):
        """Check if current index is valid"""
        valid = False
        default = None
        indexes = self.window.core.config.get("llama.idx.list")
        for item in indexes:
            if item["id"] == self.current_idx:
                valid = True
            if default is None:
                default = item["id"]
        if not valid and self.current_idx != "-":
            self.current_idx = default

    def reload(self):
        """Reload indexes"""
        self.window.core.idx.indexing.reload_loaders()
        self.update_tab_web()
        self.window.ui.nodes["tool.indexer.idx"].set_keys(self.window.controller.config.placeholder.apply_by_id("idx"))
        self.set_current_idx(self.current_idx)
        self.window.ui.nodes["tool.indexer.idx"].set_value(self.current_idx)

    def refresh(self, check: bool = True):
        """
        Refresh dialog window

        :param check: check if index is valid
        """
        if check:
            self.check_current_idx()
        self.update_tabs()

    def on_reload(self):
        """On app profile reload"""
        self.reload()

    def update_tabs(self):
        """Update tabs"""
        self.window.ui.nodes["tool.indexer.provider"].setText(self.window.core.config.get("llama.idx.storage"))
        self.update_tab_ctx()
        self.update_tab_files()
        self.update_tab_browse()

    def update_tab_ctx(self):
        """Update context tab"""
        # get database
        db = self.window.core.db.get_db()

        # last meta ID
        last_meta_id = "-"
        query = "SELECT meta_id, updated_ts FROM idx_ctx WHERE idx=:idx AND store=:store ORDER BY meta_id DESC LIMIT 1"
        stmt = text(query).bindparams(idx=self.current_idx, store=self.window.core.config.get("llama.idx.storage"))
        with db.connect() as conn:
            results = conn.execute(stmt).fetchall()
            if results:
                row = results[0]._asdict()
                last_meta_id = str(row["meta_id"]) + " ({})".format(
                    datetime.datetime.fromtimestamp(row["updated_ts"]).strftime('%Y-%m-%d %H:%M:%S'))
        self.window.ui.nodes['tool.indexer.ctx.last_meta_id'].setText(last_meta_id)

        # last meta ts
        last_meta_ts = "-"
        query = "SELECT meta_id, updated_ts FROM idx_ctx WHERE idx=:idx AND store=:store ORDER BY updated_ts DESC LIMIT 1"
        stmt = text(query).bindparams(idx=self.current_idx, store=self.window.core.config.get("llama.idx.storage"))
        with db.connect() as conn:
            results = conn.execute(stmt).fetchall()
            if results:
                row = results[0]._asdict()
                last_meta_ts = str(row["meta_id"]) + " ({})".format(
                    datetime.datetime.fromtimestamp(row["updated_ts"]).strftime('%Y-%m-%d %H:%M:%S'))
        self.window.ui.nodes['tool.indexer.ctx.last_meta_ts'].setText(last_meta_ts)

        # auto-index enabled
        auto_str = trans("tool.indexer.tab.ctx.auto.no")
        if self.window.core.config.get("llama.idx.auto"):
            auto_str = trans("tool.indexer.tab.ctx.auto.yes") + " ({})".format(self.window.core.config.get("llama.idx.auto.index"))
        self.window.ui.nodes['tool.indexer.ctx.auto_enabled'].setText(auto_str)

        # last auto-index TS
        last_str = trans('settings.llama.extra.db.never')
        if self.window.core.config.has('llama.idx.db.last'):
            last_ts = int(self.window.core.config.get('llama.idx.db.last'))
            if last_ts > 0:
                last_str = datetime.datetime.fromtimestamp(last_ts).strftime('%Y-%m-%d %H:%M:%S')
        self.window.ui.nodes['tool.indexer.ctx.last_auto'].setText(last_str)

    def update_tab_files(self) -> List[str]:
        """
        Update files tab

        :return: list of available file loaders
        """
        list_file = []
        list_file.append("txt (plain-text files)")
        providers = self.window.core.idx.indexing.get_data_providers()
        for id in providers:
            loader = providers[id]
            if "file" in loader.type:
                for ext in loader.extensions:
                    list_file.append(ext)
        list_file = sorted(list_file)
        files_str = ", ".join(list_file)
        info = trans("tool.indexer.loaders") + ": " + files_str
        self.window.ui.nodes['tool.indexer.file.loaders'].setText(info)
        return list_file

    def update_tab_browse(self):
        """Update browse tab"""
        self.window.ui.nodes['tool.indexer.browser'].update_table_view()

    def update_tab_web(self):
        """Update web tab"""
        loaders = self.window.core.idx.indexing.get_external_config()
        for loader in loaders:
            params = loaders[loader]
            for k in params:
                key_path = "tool.indexer.web.loader.config." + loader + "." + k
                value = ""
                try:
                    if params[k]["value"] is not None:
                        if params[k]["type"] == "list" and isinstance(params[k]["value"], list):
                            value = ", ".join(params[k]["value"])
                        elif params[k]["type"] == "dict" and isinstance(params[k]["value"], dict):
                            value = json.dumps(params[k]["value"])
                        else:
                            value = str(params[k]["value"])
                except Exception as e:
                    self.window.core.debug.log(e)
                if key_path in self.window.ui.nodes:
                    self.window.ui.nodes[key_path].setText(value)
                    self.window.ui.nodes[key_path].value = value

    def idx_ctx_db_all(self):
        """Index all context data"""
        # self.clear_log()
        self.window.controller.idx.indexer.index_ctx_from_ts(self.current_idx, 0)

    def idx_ctx_db_update(self):
        """Index context data from last update"""
        # self.clear_log()
        self.window.controller.idx.indexer.index_ctx_current(self.current_idx)

    def index_data(self, force: bool = False):
        """
        Index data

        :param force: force indexing
        """
        if self.current_idx == "-" or self.current_idx is None or self.current_idx == "_":
            self.window.ui.dialogs.alert(trans("tool.indexer.alert.no_idx"))
            return
        tab = self.window.ui.tabs['tool.indexer'].currentIndex()
        # self.clear_log()
        if tab == 0:
            self.index_files(force)
        elif tab == 1:
            self.index_web(force)

    def index_files(self, force: bool = False):
        """
        Index files

        :param force: force indexing
        """
        is_recursive = self.window.ui.nodes["tool.indexer.file.options.recursive"].isChecked()
        is_replace = self.window.ui.nodes["tool.indexer.file.options.replace"].isChecked()
        dir = self.window.ui.nodes["tool.indexer.file.path_dir"].value
        files = self.window.ui.nodes["tool.indexer.file.path_file"].value
        paths = []
        if dir:
            paths.append(dir)
        if files:
            paths.extend(files)
        if len(paths) == 0:
            self.window.ui.dialogs.alert(trans("tool.indexer.alert.no_files"))
            return
        if not force:
            self.window.ui.dialogs.confirm(
                type="idx.tool.index",
                id=0,
                msg=trans("tool.indexer.confirm.idx"),
            )
            return
        self.window.controller.idx.indexer.index_paths(
            paths,
            self.current_idx,
            is_replace,
            is_recursive,
        )

    def index_web(self, force: bool = False):
        """
        Index web data

        :param force: force indexing
        """
        result, loader, input_params, input_config = self.window.core.idx.ui.loaders.handle_options(
            self.window.ui.nodes["tool.indexer.web.loader"],
            "tool.indexer.web.loader.option",
            "tool.indexer.web.loader.config",
        )
        if not result:
            self.window.ui.dialogs.alert(trans("tool.indexer.alert.no_loader"))
            return

        is_replace = self.window.ui.nodes["tool.indexer.web.options.replace"].isChecked()
        if not force:
            self.window.ui.dialogs.confirm(
                type="idx.tool.index",
                id=0,
                msg=trans("tool.indexer.confirm.idx"),
            )
            return

        self.window.controller.idx.indexer.index_web(
            self.current_idx,
            loader,
            input_params,
            input_config,
            is_replace,
        )

    def on_finish_files(self):
        """On finish indexing files"""
        if self.window.ui.nodes["tool.indexer.file.options.clear"].isChecked():
            self.window.ui.nodes["tool.indexer.file.path_file"].clear()
            self.window.ui.nodes["tool.indexer.file.path_dir"].clear()
        print("Indexing files finished")

    def on_finish_web(self):
        """On finish indexing web"""
        print("Indexing web finished")

    def truncate_idx(self):
        """Truncate index"""
        self.window.controller.idx.indexer.clear(self.current_idx)

    def delete_db_idx(self, id: int, force: bool = False):
        """
        Delete index from database

        :param id: index id
        :param force: force deletion
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type="idx.tool.truncate",
                id=id,
                msg=trans("tool.indexer.confirm.remove"),
            )
            return
            
        table = self.window.ui.nodes['tool.indexer.browser'].get_current_table()
        query = "SELECT doc_id FROM {} WHERE id=:id".format(table)
        stmt = text(query).bindparams(id=id)
        with self.window.core.db.get_db().connect() as conn:
            results = conn.execute(stmt).fetchall()
            if results:
                result = results[0]._asdict()
                doc_id = result["doc_id"]

        # remove from db
        query = "DELETE FROM {} WHERE id=:id".format(table)
        stmt = text(query).bindparams(id=id)
        with self.window.core.db.get_db().begin() as conn:
            conn.execute(stmt)

        # remove from index
        if doc_id:
            self.window.core.idx.remove_doc(self.current_idx, doc_id)

        self.refresh()

    @Slot(object)
    def handle_log(self, data: Any):
        """
        Handle log message

        :param data: message to log
        """
        self.log(data)

    def log(self, data: Any):
        """
        Log message to console or logger window

        :param data: text to log
        """
        if not self.opened or data is None or str(data).strip() == "":
            return
        data = datetime.datetime.now().strftime('%H:%M:%S.%f') + ': ' + str(data)
        cur = self.window.ui.nodes["tool.indexer.status"].textCursor()  # Move cursor to end of text
        cur.movePosition(QTextCursor.End)
        s = str(data) + "\n"
        while s:
            head, sep, s = s.partition("\n")  # Split line at LF
            cur.insertText(head)  # Insert text at cursor
            if sep:  # New line if LF
                cur.insertText("\n")
        self.window.ui.nodes["tool.indexer.status"].setTextCursor(cur)  # Update visible cursor

    def clear_log(self):
        """Clear log"""
        self.window.ui.nodes["tool.indexer.status"].clear()

    def get_tables(self) -> Dict[str, Dict]:
        """
        Get tables configuration

        :return: Tables dictionary
        """
        columns = {}
        columns["idx_ctx"] = [
            'id',
            'meta_id',
            'doc_id',
            'updated_ts',
        ]
        columns["idx_external"] = [
            'id',
            'content',
            'type',
            'doc_id',
            'updated_ts',
        ]
        columns["idx_file"] = [
            'id',
            'name',
            'path',
            'doc_id',
            'updated_ts',
        ]

        tables = {
            'idx_ctx': {
                'columns': columns["idx_ctx"],
                'sort_by': columns["idx_ctx"],
                'search_fields': ['id', 'doc_id', 'meta_id'],
                'timestamp_columns': ['updated_ts'],
                'json_columns': [],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
            'idx_file': {
                'columns': columns["idx_file"],
                'sort_by': columns["idx_file"],
                'search_fields': ['id', 'doc_id', 'name', 'path'],
                'timestamp_columns': ['updated_ts'],
                'json_columns': [],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
            'idx_external': {
                'columns': columns["idx_external"],
                'sort_by': columns["idx_external"],
                'search_fields': ['id', 'doc_id', 'content', 'type'],
                'timestamp_columns': ['updated_ts'],
                'json_columns': [],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
        }
        return tables

    def setup_menu(self) -> Dict[str, QAction]:
        """
        Setup main menu (Tools)

        :return dict with menu actions
        """
        actions = {}
        actions["indexer"] = QAction(
            QIcon(":/icons/db.svg"),
            trans("tool.indexer"),
            self.window,
            checkable=False,
        )
        actions["indexer"].triggered.connect(
            lambda: self.toggle()
        )
        return actions

    def setup_dialogs(self):
        """Setup dialogs (static)"""
        self.dialog = DialogBuilder(self.window)
        self.dialog.setup()

    def get_lang_mappings(self) -> Dict[str, Dict]:
        """
        Get language mappings

        :return: dict with language mappings
        """
        return {
            'menu.text': {
                'tools.indexer': 'tool.indexer',  # menu key => translation key
            }
        }





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.26 19:00:00                  #
# ================================================== #

import os

from PySide6.QtCore import Slot

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.provider.audio_input.base import BaseProvider
from pygpt_net.core.events import Event, KernelEvent, RenderEvent
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans

from .config import Config
from .worker import Worker
from .simple import Simple


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.handler_simple = Simple(self)
        self.id = "audio_input"
        self.name = "Audio Input"
        self.type = [
            'audio.input',
        ]
        self.description = "Enables speech recognition"
        self.prefix = "Audio Input"
        self.input_text = None
        self.speech_enabled = False
        self.thread_started = False
        self.listening = False
        self.waiting = False
        self.stop = False
        self.magic_word_detected = False
        self.is_first_adjust = True
        self.empty_phrases = [
            'Thank you for watching',
        ]  # phrases to ignore (fix for empty phrases)
        self.order = 1
        self.use_locale = True
        self.input_file = "input.wav"
        self.config = Config(self)

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)
        self.init_provider()

    def init_provider(self):
        """Initialize provider options"""
        providers = self.get_providers()
        for id in providers:
            providers[id].init(self)

    def get_providers(self) -> dict:
        """
        Get audio input providers

        :return: providers dict
        """
        return self.window.core.audio.get_providers("input")

    def get_provider_options(self) -> list:
        """
        Get provider options

        :return: list of provider options
        """
        options = []
        providers = self.get_providers()
        for id in providers:
            options.append({id: providers[id].name})
        return options

    def init_tabs(self) -> dict:
        """
        Initialize provider tabs

        :return: dict of tabs
        """
        tabs = {}
        tabs["general"] = "General"
        providers = self.get_providers()
        for id in providers:
            tabs[id] = providers[id].name
        return tabs

    def setup_ui(self):
        """
        Setup UI
        """
        pass

    def is_advanced(self) -> bool:
        """
        Check if advanced options are enabled

        :return: True if advanced options are enabled
        """
        return self.get_option_value("advanced")

    def get_words(self, key: str) -> list:
        """
        Get and parse words from option string

        :param key: option key
        :return: words list
        """
        words_str = self.get_option_value(key)
        words = []
        if words_str is not None and len(words_str) > 0 and words_str.strip() != ' ':
            words = words_str.split(',')
            words = [x.strip() for x in words]  # remove white-spaces
        return words

    def toggle_recording_simple(self):
        """
        Event: AUDIO_INPUT_RECORD_TOGGLE

        Toggle recording
        """
        self.handler_simple.toggle_recording()

    def toggle_speech(self, state: bool):
        """
        Event: AUDIO_INPUT_TOGGLE

        Toggle speech

        :param state: state to set
        """
        if not self.is_advanced():
            return

        self.speech_enabled = state
        self.window.ui.plugin_addon['audio.input'].btn_toggle.setChecked(state)

        # Start thread if not started
        if state:
            self.listening = True
            self.stop = False
            self.handle_thread()
        else:
            self.listening = False
            self.set_status('')

    def attach(self, window):
        """
        Attach window

        :param window: Window instance
        """
        self.window = window

        # register provider tabs
        self.tabs = self.init_tabs()

        # register options
        self.init_options()

    def get_provider(self) -> BaseProvider:
        """
        Get audio input provider

        :return: provider instance
        """
        current = self.get_option_value("provider")
        providers = self.get_providers()
        if current not in providers:
            raise Exception("Provider '{}' not found!".format(current))
        return providers[current]

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: args
        :param kwargs: kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == Event.INPUT_BEFORE:
            self.on_input_before(data['value'])

        elif name == Event.CTX_BEGIN:
            self.on_ctx_begin(ctx)

        elif name == Event.CTX_END:
            self.on_ctx_end(ctx)

        elif name == Event.ENABLE:
            if data['value'] == self.id:
                self.on_enable()

        elif name == Event.DISABLE:
            if data['value'] == self.id:
                self.on_disable()

        elif name == Event.AUDIO_INPUT_TOGGLE:
            self.toggle_speech(data['value'])

        elif name == Event.AUDIO_INPUT_RECORD_TOGGLE:
            self.toggle_recording_simple()

        elif name == Event.AUDIO_INPUT_STOP:
            self.on_stop()

        elif name == Event.AUDIO_INPUT_TRANSCRIBE:
            self.handle_transcribe(data["path"])

        elif name == Event.PLUGIN_OPTION_GET:
            if "name" in data and data["name"] == "audio.input.advanced":
                data["value"] = self.is_advanced()

    def on_ctx_begin(self, ctx: CtxItem):
        """
        Event: CTX_BEGIN

        :param ctx: CtxItem
        """
        self.waiting = True

    def on_ctx_end(self, ctx: CtxItem):
        """
        Event: CTX_END

        :param ctx: CtxItem
        """
        self.waiting = False

    def on_enable(self):
        """Event: ENABLE"""
        if not self.is_advanced():
            return

        self.speech_enabled = True
        self.handle_thread()

    def on_disable(self):
        """Event: DISABLE"""
        self.speech_enabled = False
        self.listening = False
        self.stop = True
        self.window.ui.plugin_addon['audio.input'].btn_toggle.setChecked(False)
        self.set_status('')
        self.window.update_status("")

    def on_stop(self):
        """Event: AUDIO_INPUT_STOP"""
        if not self.is_advanced():
            return

        self.stop = True
        self.listening = False
        self.speech_enabled = False
        self.set_status('')
        self.window.update_status("")

    def on_input_before(self, text: str):
        """
        Event: INPUT_BEFORE

        :param text: text from input
        """
        self.input_text = text

    def destroy(self):
        """Destroy thread"""
        self.waiting = True
        self.listening = False
        self.thread_started = False
        self.set_status('')

    @Slot(str, str)
    def handle_transcribed(self, path: str, text: str):
        """
        Handle transcribed text

        :param path: audio file path
        :param text: transcribed text
        """
        self.window.tools.get("transcriber").on_transcribe(path, text)

    def handle_transcribe(self, path: str):
        """
        Handle transcribe thread

        :param path: audio file path
        """
        try:
            worker = Worker()
            worker.from_defaults(self)
            worker.path = path
            worker.transcribe = True  # file transcribe mode

            # signals
            worker.signals.transcribed.connect(self.handle_transcribed)
            worker.signals.finished.connect(self.handle_input)
            worker.signals.destroyed.connect(self.handle_destroy)
            worker.signals.started.connect(self.handle_started)
            worker.signals.stopped.connect(self.handle_stop)

            worker.run_async()

        except Exception as e:
            self.error(e)

    def handle_thread(self, force: bool = False):
        """
        Handle listener thread

        :param force: force start
        """
        if self.thread_started and not force:
            return

        try:
            worker = Worker()
            worker.from_defaults(self)
            worker.path = os.path.join(
                self.window.core.config.path,
                self.input_file,
            )
            worker.advanced = self.is_advanced()  # advanced mode

            # signals
            worker.signals.transcribed.connect(self.handle_transcribed)
            worker.signals.finished.connect(self.handle_input)
            worker.signals.destroyed.connect(self.handle_destroy)
            worker.signals.started.connect(self.handle_started)
            worker.signals.stopped.connect(self.handle_stop)
            worker.signals.on_realtime.connect(self.handle_realtime)

            worker.run_async()

        except Exception as e:
            self.error(e)

    def can_listen(self) -> bool:
        """
        Check if can listen

        :return: True if can listen
        """
        state = True
        if self.get_option_value('wait_response') \
                and self.window.controller.chat.input.generating:
            state = False
        if self.window.controller.chat.input.locked:
            state = False
        return state

    def set_status(self, status: str):
        """
        Set status

        :param status: status text
        """
        self.window.ui.plugin_addon['audio.input'].set_status(status)

    @Slot(str)
    def handle_realtime(self, path: str):
        """
        Handle realtime audio input

        :param path: audio input file path
        """
        self.window.controller.chat.audio.handle_input(path)

    @Slot(object, object)
    def handle_input(self, text: str, ctx: CtxItem = None):
        """
        Insert text to input and send

        :param text: text
        :param ctx: CtxItem
        """
        if text is None or text.strip() == '':
            return

        # simple mode
        if not self.is_advanced():
            check = text.strip().lower()
            for phrase in self.empty_phrases:
                phrase_check = phrase.strip().lower()
                if phrase_check in check:
                    return

        # advanced mode
        if self.is_advanced():
            if not self.can_listen():
                return

            # check prefix words
            prefix_words = self.get_words('prefix_words')
            if len(prefix_words) > 0:
                for word in prefix_words:
                    check_text = text.lower().strip()
                    check_word = word.lower().strip()
                    if not check_text.startswith(check_word):
                        self.window.update_status(trans('audio.speak.ignoring'))
                        self.set_status(trans('audio.speak.ignoring'))
                        return

            # previous magic word detected state
            magic_prev_detected = self.magic_word_detected
            # save the magic word detected state before checking for magic word

            # check for magic word
            is_magic_word = False
            if self.get_option_value('magic_word'):
                for word in self.get_words('magic_words'):
                    # prepare magic word
                    check_word = word.lower().replace('.', '')
                    check_text = text.lower()
                    if check_text.startswith(check_word):
                        is_magic_word = True
                        self.set_status(trans('audio.magic_word.detected'))
                        self.window.update_status(trans('audio.magic_word.detected'))
                        break

            # if magic word enabled
            if self.get_option_value('magic_word'):
                if not is_magic_word:
                    if self.get_option_value('magic_word_reset'):
                        self.magic_word_detected = False
                else:
                    self.magic_word_detected = True

                # if not previously detected, then abort now
                if not magic_prev_detected:
                    if not is_magic_word:
                        self.window.update_status(trans('audio.magic_word.invalid'))
                    self.window.ui.nodes['input'].setText(text)
                    return
                else:
                    self.window.update_status("")

        # update input text
        self.window.ui.nodes['input'].setText(text)

        # send text
        if self.get_option_value('auto_send'):
            # to: notepad
            if self.window.controller.notepad.is_active():
                idx = self.window.controller.notepad.get_current_active()
                self.window.controller.notepad.append_text(text, idx)
                self.set_status('')

                data = {}
                event = RenderEvent(RenderEvent.CLEAR_INPUT, data)
                self.window.dispatch(event)  # clear here

            # to: calendar
            elif self.window.controller.calendar.is_active():
                self.window.controller.calendar.note.append_text(text)
                self.set_status('')

                data = {}
                event = RenderEvent(RenderEvent.CLEAR_INPUT, data)
                self.window.dispatch(event)  # clear here

            # to: chat
            else:
                self.set_status('...')
                self.window.update_status(trans('audio.speak.sending'))
                prefix = ""
                if self.window.controller.agent.legacy.enabled():
                    prefix = "user: "

                context = BridgeContext()
                context.prompt = prefix + text
                extra = {}
                event = KernelEvent(KernelEvent.INPUT_SYSTEM, {
                    'context': context,
                    'extra': extra,
                })
                self.window.dispatch(event)  # send text, input clear in send method
                self.set_status('')

    @Slot(object)
    def handle_status(self, data: str):
        """
        Handle thread status msg signal

        :param data: message
        """
        self.set_status(str(data))
        self.window.update_status(str(data))

    @Slot()
    def handle_destroy(self):
        """Handle listener destroyed"""
        self.thread_started = False
        self.set_status('')
        self.window.update_status("")

    @Slot()
    def handle_started(self):
        """Handle listening started"""
        self.thread_started = True
        # print("Whisper is listening...")

    @Slot()
    def handle_stop(self):
        """Handle stop listening"""
        self.thread_started = False
        self.listening = False
        self.stop = False
        self.window.update_status("")
        self.set_status('')
        self.toggle_speech(False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

from packaging.version import parse as parse_version, Version


class Patch:
    def __init__(self, window=None):
        self.window = window

    def execute(self, version: Version) -> bool:
        """
        Migrate to current app version

        :param version: current app version
        :return: True if migrated
        """
        data = self.window.core.models.items
        updated = False

        # get version of models config
        current = self.window.core.models.get_version()
        old = parse_version(current)

        # check if models file is older than current app version
        is_old = False
        if old < version:
            is_old = True

            # < 0.9.1
            if old < parse_version("0.9.1"):
                # apply meta only (not attached in 0.9.0)
                print("Migrating models from < 0.9.1...")
                updated = True

            # < 2.0.1
            if old < parse_version("2.0.1"):
                print("Migrating models from < 2.0.1...")
                self.window.core.updater.patch_file('models.json', True)  # force replace file
                self.window.core.models.load()
                data = self.window.core.models.items
                updated = True

            # < 2.0.96  <--- patch for llama-index modes
            if old < parse_version("2.0.96"):
                print("Migrating models from < 2.0.96...")
                self.window.core.updater.patch_file('models.json', True)  # force replace file
                self.window.core.models.load()
                data = self.window.core.models.items
                updated = True

            # < 2.0.105  <--- patch for llama-index gpt4-turbo
            if old < parse_version("2.0.105"):
                print("Migrating models from < 2.0.105...")
                self.window.core.updater.patch_file('models.json', True)  # force replace file
                self.window.core.models.load()
                data = self.window.core.models.items
                updated = True

            '''
            # < 2.0.104  <--- patch to new format
            if old < parse_version("2.0.104"):
                print("Migrating models from < 2.0.104...")
                for id in data:
                    model = data[id]
                    dict_name = model.to_dict()
                    model.from_dict(dict_name)

                    # patch missing llama_index provider
                    if "llama_index" in model.mode:
                        if model.id.startswith("gpt-") or model.id.startswith("text-davinci-"):
                            model.llama_index["provider"] = "openai"
                            if model.id.startswith("gpt-"):
                                model.llama_index['mode'] = ["chat"]
                            model.llama_index['args'] = [
                                {
                                    "name": "model_name",
                                    "value": model.id,
                                    "type": "str",
                                }
                            ]
                            model.llama_index['env'] = [
                                {
                                    "name": "OPENAI_API_KEY",
                                    "value": "{api_key}",
                                }
                            ]
                    if "langchain" in model.mode:
                        if model.id.startswith("gpt-") or model.id.startswith("text-davinci-"):
                            model.langchain['args'] = [
                                {
                                    "name": "model_name",
                                    "value": model.id,
                                    "type": "str",
                                }
                            ]
                            model.langchain['env'] = [
                                {
                                    "name": "OPENAI_API_KEY",
                                    "value": "{api_key}",
                                }
                            ]
                updated = True
            '''

            # < 2.0.107  <--- patch for deprecated davinci, replace with gpt-3.5-turbo-instruct
            if old < parse_version("2.0.107"):
                print("Migrating models from < 2.0.107...")
                if "text-davinci-002" in data:
                    del data["text-davinci-002"]
                if "text-davinci-003" in data:
                    data["text-davinci-003"].id = "gpt-3.5-turbo-instruct"
                    data["text-davinci-003"].name = "gpt-3.5-turbo-instruct"
                    if "llama_index" in data["text-davinci-003"].mode:
                        data["text-davinci-003"].mode.remove("llama_index")
                    if len(data["text-davinci-003"].langchain["args"]) > 0:
                        if data["text-davinci-003"].langchain["args"][0]["name"] == "model_name":
                            data["text-davinci-003"].langchain["args"][0]["value"] = "gpt-3.5-turbo-instruct"
                    data["text-davinci-003"].llama_index["args"] = []
                    data["text-davinci-003"].llama_index["env"] = []
                    data["text-davinci-003"].llama_index["provider"] = None
                    # replace "text-davinci-003" with "gpt-3.5-turbo-instruct"
                    if "gpt-3.5-turbo-instruct" not in data:
                        data["gpt-3.5-turbo-instruct"] = data["text-davinci-003"]
                        del data["text-davinci-003"]
                updated = True

            # < 2.0.123  <--- update names to models IDs
            if old < parse_version("2.0.123"):
                print("Migrating models from < 2.0.123...")
                if "gpt-4-1106-preview" in data:
                    data["gpt-4-1106-preview"].name = "gpt-4-1106-preview"
                if "gpt-4-vision-preview" in data:
                    data["gpt-4-vision-preview"].name = "gpt-4-vision-preview"
                updated = True

            # < 2.0.134  <--- add agent mode
            if old < parse_version("2.0.134"):
                print("Migrating models from < 2.0.134...")
                exclude = ["gpt-3.5-turbo-instruct", "gpt-4-vision-preview"]
                for id in data:
                    model = data[id]
                    if model.id.startswith("gpt-") and model.id not in exclude:
                        if "agent" not in model.mode:
                            model.mode.append("agent")
                updated = True

            # fix typo in gpt-4 turbo preview for llama
            if old < parse_version("2.1.15"):
                print("Migrating models from < 2.1.15...")
                if "gpt-4-turbo-preview" in data:
                    data["gpt-4-turbo-preview"].llama_index["args"] = [
                        {
                            "name": "model",
                            "value": "gpt-4-turbo-preview",
                            "type": "str",
                        }
                    ]
                updated = True

            # add API endpoint
            if old < parse_version("2.1.19"):
                print("Migrating models from < 2.1.19...")
                for id in data:
                    model = data[id]
                    if model.id.startswith("gpt-"):
                        if "env" not in model.llama_index:
                            model.llama_index["env"] = []
                        is_endpoint = False
                        for arg in model.llama_index["env"]:
                            if "OPENAI_API_BASE" in arg["name"]:
                                is_endpoint = True
                                break
                        if not is_endpoint:
                            model.llama_index["env"].append(
                                {
                                    "name": "OPENAI_API_BASE",
                                    "value": "{api_endpoint}",
                                }
                            )
                        if "env" not in model.langchain:
                            model.langchain["env"] = []
                        is_endpoint = False
                        for arg in model.langchain["env"]:
                            if "OPENAI_API_BASE" in arg["name"]:
                                is_endpoint = True
                                break
                        if not is_endpoint:
                            model.langchain["env"].append(
                                {
                                    "name": "OPENAI_API_BASE",
                                    "value": "{api_endpoint}",
                                }
                            )
                updated = True

            if old < parse_version("2.1.45"):
                print("Migrating models from < 2.1.45...")
                # add missing 2024-04-09
                updated = True

            if old < parse_version("2.2.6"):
                print("Migrating models from < 2.2.6...")
                # add missing gpt-4-turbo
                updated = True

            # < 2.2.7  <--- add expert mode
            if old < parse_version("2.2.7"):
                print("Migrating models from < 2.2.7...")
                exclude = ["gpt-3.5-turbo-instruct", "gpt-4-vision-preview"]
                for id in data:
                    model = data[id]
                    if model.id.startswith("gpt-") and model.id not in exclude:
                        if "expert" not in model.mode:
                            model.mode.append("expert")
                updated = True

            # < 2.2.19  <--- add gpt-4o
            if old < parse_version("2.2.19"):
                print("Migrating models from < 2.2.19...")
                # add gpt-4o
                updated = True

            # < 2.2.20  <--- add gpt-4o-mini
            if old < parse_version("2.2.20"):
                print("Migrating models from < 2.2.20...")
                # add gpt-4o-mini
                updated = True

            # < 2.2.22  <--- add Llama index models
            if old < parse_version("2.2.22"):
                print("Migrating models from < 2.2.22...")
                # add Gemini, Claude, Llama3, Mistral and etc.
                updated = True

            # < 2.2.28  <--- add Llama index models
            if old < parse_version("2.2.28"):
                print("Migrating models from < 2.2.28...")
                # add Llama3.1 70b and 405b, mistral-large
                updated = True

            # < 2.2.33  <--- add agent and expert modes
            if old < parse_version("2.2.33"):
                print("Migrating models from < 2.2.33...")
                exclude = ["dall-e-2", "dall-e-3", "gpt-3.5-turbo-instruct"]
                for id in data:
                    model = data[id]
                    if model.id not in exclude:
                        if "agent" not in model.mode:
                            model.mode.append("agent")
                        if "expert" not in model.mode:
                            model.mode.append("expert")
                # change dalle model names
                if "dall-e-2" in data:
                    data["dall-e-2"].name = "dall-e-2"
                if "dall-e-3" in data:
                    data["dall-e-3"].name = "dall-e-3"
                updated = True

            # < 2.3.3  <--- add o1-preview, o1-mini, Bielik v2.2
            if old < parse_version("2.3.3"):
                print("Migrating models from < 2.3.3...")
                # add o1-preview, o1-mini, Bielik v2.2
                updated = True

            # < 2.4.0  <--- add langchain
            if old < parse_version("2.4.0"):
                print("Migrating models from < 2.4.0...")
                if 'bielik-11b-v2.2-instruct:Q4_K_M' in data:
                    model = data['bielik-11b-v2.2-instruct:Q4_K_M']
                    if "langchain" not in model.mode:
                        model.mode.append("langchain")
                updated = True

            # < 2.4.10  <--- add agent_llama mode
            if old < parse_version("2.4.10"):
                print("Migrating models from < 2.4.10...")
                exclude = ["gpt-3.5-turbo-instruct"]
                for id in data:
                    model = data[id]
                    if model.id.startswith("gpt-") and model.id not in exclude:
                        if "agent_llama" not in model.mode:
                            model.mode.append("agent_llama")
                updated = True

            # < 2.4.11  <--- add agent_llama mode to rest of models
            if old < parse_version("2.4.11"):
                print("Migrating models from < 2.4.11...")
                exclude = [
                    "gpt-3.5-turbo-instruct",
                    "dall-e-2",
                    "dall-e-3",
                    "o1-preview",
                    "o1-mini",
                ]
                for id in data:
                    model = data[id]
                    if model.id not in exclude:
                        if "agent_llama" not in model.mode:
                            model.mode.append("agent_llama")
                updated = True

            # < 2.4.34 <--- add gpt-4o-audio-preview, gpt-4o-2024-11-20
            if old < parse_version("2.4.34"):
                print("Migrating models from < 2.4.34...")
                # add missing gpt-4o-audio-preview, gpt-4o-2024-11-20
                updated = True

            # < 2.4.46  <--- add separated API keys
            if old < parse_version("2.4.46"):
                print("Migrating models from < 2.4.46...")
                azure_endpoint = ""
                azure_api_version = ""
                google_key = ""
                anthropic_key = ""
                for id in data:
                    model = data[id]
                    # OpenAI
                    if model.id.startswith("gpt-") or model.id.startswith("o1-"):
                        # langchain
                        is_endpoint = False
                        is_version = False
                        for item in model.langchain["env"]:
                            if item["name"] == "AZURE_OPENAI_ENDPOINT":
                                is_endpoint = True
                                if (item["value"]
                                        and item["value"] not in ["{api_azure_endpoint}", "{api_endpoint}"]):
                                    azure_endpoint = item["value"]
                                item["value"] = "{api_azure_endpoint}"
                            elif item["name"] == "OPENAI_API_VERSION":
                                is_version = True
                                if (item["value"]
                                        and item["value"] not in ["{api_azure_version}"]):
                                    azure_api_version = item["value"]
                                item["value"] = "{api_azure_version}"
                        if not is_endpoint:
                            model.langchain["env"].append(
                                {
                                    "name": "AZURE_OPENAI_ENDPOINT",
                                    "value": "{api_azure_endpoint}",
                                }
                            )
                        if not is_version:
                            model.langchain["env"].append(
                                {
                                    "name": "OPENAI_API_VERSION",
                                    "value": "{api_azure_version}",
                                }
                            )

                        # llama
                        is_endpoint = False
                        is_version = False
                        for item in model.llama_index["env"]:
                            if item["name"] == "AZURE_OPENAI_ENDPOINT":
                                is_endpoint = True
                                if (item["value"]
                                        and item["value"] not in ["{api_azure_endpoint}", "{api_endpoint}"]):
                                    azure_endpoint = item["value"]
                                item["value"] = "{api_azure_endpoint}"
                            elif item["name"] == "OPENAI_API_VERSION":
                                is_version = True
                                if (item["value"]
                                        and item["value"] not in ["{api_azure_version}"]):
                                    azure_api_version = item["value"]
                                item["value"] = "{api_azure_version}"
                        if not is_endpoint:
                            model.llama_index["env"].append(
                                {
                                    "name": "AZURE_OPENAI_ENDPOINT",
                                    "value": "{api_azure_endpoint}",
                                }
                            )
                        if not is_version:
                            model.llama_index["env"].append(
                                {
                                    "name": "OPENAI_API_VERSION",
                                    "value": "{api_azure_version}",
                                }
                            )

                    # Anthropic
                    elif model.id.startswith("claude-"):
                        is_key = False
                        for item in model.langchain["env"]:
                            if item["name"] == "ANTHROPIC_API_KEY":
                                is_key = True
                                if (item["value"]
                                        and item["value"] not in ["{api_key}"]):
                                    anthropic_key = item["value"]
                                item["value"] = "{api_key_anthropic}"
                        if not is_key:
                            model.langchain["env"].append(
                                {
                                    "name": "ANTHROPIC_API_KEY",
                                    "value": "{api_key_anthropic}",
                                }
                            )
                        is_key = False
                        for item in model.llama_index["env"]:
                            if item["name"] == "ANTHROPIC_API_KEY":
                                is_key = True
                                if (item["value"]
                                        and item["value"] not in ["{api_key}"]):
                                    anthropic_key = item["value"]
                                item["value"] = "{api_key_anthropic}"
                        if not is_key:
                            model.llama_index["env"].append(
                                {
                                    "name": "ANTHROPIC_API_KEY",
                                    "value": "{api_key_anthropic}",
                                }
                            )
                    # Google
                    elif model.id.startswith("gemini-"):
                        is_key = False
                        for item in model.langchain["env"]:
                            if item["name"] == "GOOGLE_API_KEY":
                                is_key = True
                                if (item["value"]
                                        and item["value"] not in ["{api_key}"]):
                                    google_key = item["value"]
                                item["value"] = "{api_key_google}"
                        if not is_key:
                            model.langchain["env"].append(
                                {
                                    "name": "GOOGLE_API_KEY",
                                    "value": "{api_key_google}",
                                }
                            )
                        is_key = False
                        for item in model.llama_index["env"]:
                            if item["name"] == "GOOGLE_API_KEY":
                                is_key = True
                                if (item["value"]
                                        and item["value"] not in ["{api_key}"]):
                                    google_key = item["value"]
                                item["value"] = "{api_key_google}"
                        if not is_key:
                            model.llama_index["env"].append(
                                {
                                    "name": "GOOGLE_API_KEY",
                                    "value": "{api_key_google}",
                                }
                            )
                # move API keys to config
                config_updated = False
                if azure_endpoint:
                    self.window.core.config.set("api_azure_endpoint", azure_endpoint)
                    config_updated = True
                if azure_api_version:
                    self.window.core.config.set("api_azure_version", azure_api_version)
                    config_updated = True
                if google_key:
                    self.window.core.config.set("api_key_google", google_key)
                    config_updated = True
                if anthropic_key:
                    self.window.core.config.set("api_key_anthropic", anthropic_key)
                    config_updated = True
                if config_updated:
                    self.window.core.config.save()
                updated = True

            # < 2.4.47 <--- add gemini-2.0-flash-exp
            if old < parse_version("2.4.47"):
                print("Migrating models from < 2.4.47...")
                # add gemini-2.0-flash-exp
                updated = True

            # < 2.5.0 <--- add o1, DeepSeek R1, V3
            if old < parse_version("2.5.0"):
                print("Migrating models from < 2.5.0...")
                # add o1, DeepSeek R1, V3
                updated = True

            # < 2.5.2  <--- update names to models IDs
            if old < parse_version("2.5.2"):
                print("Migrating models from < 2.5.2...")
                for id in data:
                    model = data[id]
                    if model.name.startswith("DeepSeek Ollama"):
                        model.name = model.id
                updated = True

            # < 2.5.4 <--- add o3-mini, update output tokens in o1, o1-mini, o1-preview
            if old < parse_version("2.5.4"):
                print("Migrating models from < 2.5.4...")
                for id in data:
                    model = data[id]
                    if model.id == "o1":
                        model.tokens = 100000
                    elif model.id == "o1-mini":
                        model.tokens = 65536
                    elif model.id == "o1-preview":
                        model.tokens = 65536
                updated = True

            # < 2.5.8 <--- add gpt-4.5-preview and sonar models (Perplexity)
            if old < parse_version("2.5.8"):
                print("Migrating models from < 2.5.8...")
                # add gpt-4.5-preview, sonar, R1
                updated = True

        # update file
        if updated:
            data = dict(sorted(data.items()))
            self.window.core.models.items = data
            self.window.core.models.save()

            # also patch any missing models
            self.window.core.models.patch_missing()

        return updated



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 01:00:00                  #
# ================================================== #

import datetime
from typing import Optional, Tuple, List, Dict, Any

from packaging.version import Version

from pygpt_net.item.index import IndexItem
from pygpt_net.provider.core.index.base import BaseProvider
from pygpt_net.provider.core.index.json_file import JsonFileProvider
from pygpt_net.provider.core.index.db_sqlite import DbSqliteProvider
from pygpt_net.provider.vector_stores import Storage

from .indexing import Indexing
from .llm import Llm
from .chat import Chat
from .metadata import Metadata
from .ui import UI

from .types.ctx import Ctx
from .types.external import External
from .types.files import Files


class Idx:
    def __init__(self, window=None):
        """
        Indexer core

        :param window: Window instance
        """
        self.window = window
        self.indexing = Indexing(window)
        self.llm = Llm(window)
        self.storage = Storage(window)
        self.chat = Chat(window, self.storage)
        self.metadata = Metadata(window)
        self.ui = UI(window)

        self.providers = {
            "json_file": JsonFileProvider(window),  # only for patching
            "db_sqlite": DbSqliteProvider(window),
        }
        self.provider = "db_sqlite"
        self.items = {}
        self.initialized = False

        # internal types
        self.ctx = Ctx(window, self.get_provider())
        self.external = External(window, self.get_provider())
        self.files = Files(window, self.get_provider())

    def install(self):
        """Install provider data"""
        self.get_provider().install()

    def patch(self, app_version: Version) -> bool:
        """
        Patch provider data

        :param app_version: app version
        :return: True if success
        """
        return self.get_provider().patch(app_version)

    def init(self):
        """Initialize indexes"""
        if not self.initialized:
            self.load()
            self.initialized = True

    def get_current_store(self) -> str:
        """
        Get current vector store name/ID

        :return: vector store name
        """
        return self.window.core.config.get('llama.idx.storage')

    def get_provider(self) -> BaseProvider:
        """
        Get provider instance

        :return: provider instance
        """
        return self.providers.get(self.provider)

    def store_index(self, idx: str = "base"):
        """
        Store (persist) index data

        :param idx: index name/ID
        """
        self.storage.store(idx)

    def remove_index(
            self,
            idx: str = "base",
            truncate: bool = False
    ) -> bool:
        """
        Truncate index

        :param idx: index name
        :param truncate: truncate index (remove all data)
        :return: True if success
        """
        # get current store
        store = self.get_current_store()

        # clear db data
        self.ctx.truncate(store, idx)
        self.files.truncate(store, idx)

        # clear ctx data indexed status
        self.window.core.ctx.idx.truncate_indexed(store, idx)

        # clear storage, remove index
        if truncate:
            return self.storage.truncate(idx)
        else:
            return self.storage.remove(idx)

    def index_files(
            self,
            idx: str = "base",
            path: Optional[str] = None,
            replace: Optional[bool] = None,
            recursive: Optional[bool] = None,
    ) -> Tuple[Dict, List[str]]:
        """
        Index file or directory of files

        :param idx: index name
        :param path: path to file or directory
        :param replace: replace index
        :param recursive: recursive indexing
        :return: dict with indexed files (path -> id), list with errors
        """
        llm, embed_model = self.llm.get_service_context()
        index = self.storage.get(
            id=idx,
            llm=llm,
            embed_model=embed_model,
        )  # get or create index
        files, errors = self.indexing.index_files(
            idx=idx,
            index=index,
            path=path,
            replace=replace,
            recursive=recursive,
        )  # index files
        if len(files) > 0:
            self.storage.store(
                id=idx,
                index=index,
            )  # store index

        if errors:
            self.log("Error: " + str(errors))
        return files, errors

    def index_db_by_meta_id(
            self,
            idx: str = "base",
            id: int = 0,
            from_ts: int = 0
    ) -> Tuple[int, List[str]]:
        """
        Index records from db by meta id

        :param idx: index name
        :param id: CtxMeta id
        :param from_ts: timestamp from
        :return: num of indexed files, list with errors
        """
        llm, embed_model = self.llm.get_service_context()
        index = self.storage.get(
            id=idx,
            llm=llm,
            embed_model=embed_model,
        )  # get or create index
        num, errors = self.indexing.index_db_by_meta_id(
            idx=idx,
            index=index,
            id=id,
            from_ts=from_ts,
        )  # index db records
        if num > 0:
            self.storage.store(
                id=idx,
                index=index,
            )  # store index

        if errors:
            self.log("Error: " + str(errors))
        return num, errors

    def index_db_from_updated_ts(
            self,
            idx: str = "base",
            from_ts: int = 0
    ) -> Tuple[int, List[str]]:
        """
        Index records from db by meta id

        :param idx: index name
        :param from_ts: timestamp from
        :return: num of indexed files, list with errors
        """
        llm, embed_model = self.llm.get_service_context()
        index = self.storage.get(
            id=idx,
            llm=llm,
            embed_model=embed_model,
        )  # get or create index
        num, errors = self.indexing.index_db_from_updated_ts(
            idx=idx,
            index=index,
            from_ts=from_ts,
        )  # index db records
        if num > 0:
            self.storage.store(
                id=idx,
                index=index,
            )  # store index

        if errors:
            self.log("Error: " + str(errors))
        return num, errors

    def index_urls(
            self,
            idx: str = "base",
            urls: Optional[List[str]] = None,
            type: str = "webpage",
            extra_args: Optional[Dict[str, Any]] = None
    ) -> Tuple[int, List[str]]:
        """
        Index URLs

        :param idx: index name
        :param urls: list of urls
        :param type: type of url
        :param extra_args: extra args
        :return: num of indexed, list with errors
        """
        llm, embed_model = self.llm.get_service_context()
        index = self.storage.get(
            id=idx,
            llm=llm,
            embed_model=embed_model,
        )  # get or create index
        n, errors = self.indexing.index_urls(
            idx=idx,
            index=index,
            urls=urls,
            type=type,
            extra_args=extra_args,
        )  # index urls
        if n > 0:
            self.storage.store(
                id=idx,
                index=index,
            )  # store index

        if errors:
            self.log("Error: " + str(errors))
        return n, errors

    def index_web(
            self,
            idx: str = "base",
            type: str = "webpage",
            params: Optional[Dict[str, Any]] = None,
            config: Optional[Dict[str, Any]] = None,
            replace: Optional[bool] = None,
    ) -> Tuple[int, list]:
        """
        Index URLs

        :param idx: index name
        :param type: type of url
        :param params: extra args
        :param config: extra config
        :param replace: replace index
        :return: num of indexed, list with errors
        """
        # update config params
        self.indexing.update_loader_args(type, config)

        llm, embed_model = self.llm.get_service_context()
        index = self.storage.get(
            id=idx,
            llm=llm,
            embed_model=embed_model,
        )  # get or create index
        n, errors = self.indexing.index_url(
            idx=idx,
            index=index,
            url="",
            type=type,
            extra_args=params,
            is_tmp=False,
            replace=replace,
        )
        if n > 0:
            self.storage.store(
                id=idx,
                index=index,
            )  # store index

        if errors:
            self.log("Error: " + str(errors))
        return n, errors

    def get_idx_data(
            self,
            idx: Optional[str] = None
    ) -> Dict[str, Dict]:
        """
        Get indexed files data

        :param idx: index name
        :return: indexed files data (idx -> items)
        """
        indexes = {}
        store_id = self.get_current_store()
        if idx is not None:
            if store_id in self.items and idx in self.items[store_id]:
                indexes[idx] = self.items[store_id][idx].items
        else:
            # all indexes
            if store_id in self.items:
                for idx in self.items[store_id]:
                    indexes[idx] = self.items[store_id][idx].items
        return indexes

    def get_by_idx(self, idx: int) -> Optional[str]:
        """
        Return idx by list index

        :param idx: idx on list
        :return: idx name
        """
        items = self.window.core.config.get('llama.idx.list')
        if items is not None:
            if idx < len(items):
                return items[idx]['id']

    def get_idx_by_name(self, name: str) -> Optional[int]:
        """
        Return idx on list by name

        :param name: idx name
        :return: idx on list
        """
        items = self.window.core.config.get('llama.idx.list')
        if items is not None:
            for idx, item in enumerate(items):
                if item['id'] == name:
                    return idx

    def get_default_idx(self) -> Optional[str]:
        """
        Return default idx

        :return: idx name
        """
        if len(self.items) > 0:
            return self.get_by_idx(0)

    def has(self, idx: str) -> bool:
        """
        Check if index exists

        :param idx: index name
        :return: True if index exists
        """
        store_id = self.get_current_store()
        if store_id in self.items:
            return idx in self.items[store_id]
        return False

    def get(self, idx: str) -> Optional[IndexItem]:
        """
        Return index data from current storage

        :param idx: index name
        :return: IndexItem object
        """
        store_id = self.get_current_store()
        if store_id in self.items and idx in self.items[store_id]:
            return self.items[store_id][idx]

    def get_all(self) -> Dict[str, IndexItem]:
        """
        Return all indexes in current store

        :return: all indexes
        """
        store_id = self.get_current_store()
        if store_id in self.items:
            return self.items[store_id]
        return {}

    def append(
            self,
            idx: str,
            files: Dict[str, str]
    ):
        """
        Append indexed files to index

        :param idx: index name
        :param files: dict of indexed files (path -> doc_id)
        """
        # create store if not exists
        store_id = self.get_current_store()
        if store_id not in self.items:
            self.items[store_id] = {}

        # create index if not exists
        if idx not in self.items[store_id]:
            self.items[store_id][idx] = IndexItem()
            self.items[store_id][idx].id = idx
            self.items[store_id][idx].name = idx  # use index id as name

        # append indexed files
        for path in files:
            doc_id = files[path]
            file_id = self.files.get_id(path)
            ts = int(datetime.datetime.now().timestamp())
            if file_id not in self.items[store_id][idx].items:
                id = self.files.append(
                    store_id=store_id,
                    idx=idx,
                    file_id=file_id,
                    path=path,
                    doc_id=doc_id,
                )
                if id is not None:
                    self.items[store_id][idx].items[file_id] = {
                        "id": doc_id,
                        "db_id": id,  # DB id
                        "path": path,
                        "indexed_ts": ts,
                    }
            else:
                # update indexed timestamp only
                self.files.update(
                    id=self.items[store_id][idx].items[file_id]["db_id"],  # DB id
                    doc_id=doc_id,
                    ts=ts,
                )
                self.items[store_id][idx].items[file_id]["id"] = doc_id
                self.items[store_id][idx].items[file_id]["indexed_ts"] = ts

    def remove_doc(
            self,
            idx: str,
            doc_id: str
    ):
        """
        Remove document from index

        :param idx: index name (id)
        :param doc_id: document ID (in storage)
        """
        self.llm.get_service_context()  # init environment only (ENV API keys, etc.)
        if self.storage.remove_document(idx, doc_id):
            self.log("Removed document from index: " + idx + " - " + doc_id)

    def remove_file(
            self,
            idx: str,
            file: str
    ):
        """
        Remove file from index

        :param idx: index name
        :param file: file ID
        """
        self.llm.get_service_context()  # init environment only (ENV API keys, etc.)
        store_id = self.get_current_store()
        if store_id in self.items and idx in self.items[store_id]:
            if file in self.items[store_id][idx].items:
                # remove from storage
                doc_id = self.items[store_id][idx].items[file]["id"]
                self.storage.remove_document(
                    id=idx,
                    doc_id=doc_id,
                )
                # remove from index data and db
                del self.items[store_id][idx].items[file]
                self.files.remove(store_id, idx, doc_id)

    def load(self):
        """Load indexes and indexed items"""
        store_ids = self.storage.get_ids()
        for store_id in store_ids:
            self.items[store_id] = self.get_provider().load(store_id)
            # replace workdir placeholder with current workdir
            for idx in self.items[store_id]:
                for id in self.items[store_id][idx].items:
                    file = self.items[store_id][idx].items[id]
                    if 'path' in file and file['path'] is not None:
                        self.items[store_id][idx].items[id]['path'] = \
                            self.window.core.filesystem.to_workdir(file['path'])

    def sync(self):
        """Sync idx items from config"""
        items = self.window.core.config.get('llama.idx.list')
        store_id = self.get_current_store()
        if items is not None:
            if store_id not in self.items:
                self.items[store_id] = {}
            for item in items:
                idx = item['id']
                if idx not in self.items[store_id]:
                    self.items[store_id][idx] = IndexItem()
                    self.items[store_id][idx].id = idx
                    self.items[store_id][idx].name = idx
                    self.items[store_id][idx].store = store_id
                else:
                    self.items[store_id][idx].id = idx
                    self.items[store_id][idx].name = idx

    def get_idx_ids(self) -> List[Dict[str, str]]:
        """
        Get list of indexes

        :return: list of indexes
        """
        ids = []
        data =  self.window.core.config.get('llama.idx.list')
        if data is not None:
            for item in data:
                name = item['name']
                if name is None or name == "":
                    name = item['id']
                ids.append({item['id']: name})
        return ids

    def clear(self, idx: str):
        """
        Clear index items

        :param idx: index name/id
        """
        store_id = self.get_current_store()
        if store_id in self.items and idx in self.items[store_id]:
            self.items[store_id][idx].items = {}
        self.get_provider().truncate(store_id, idx)

    def get_counters(self, type: str) -> Dict[str, Dict[str, int]]:
        """
        Get counters (stats, count items by type [file, ctx, external])

        :param type: type of counter (file, ctx, external)
        :return: dict of counters: [store][idx] -> count
        """
        return self.get_provider().get_counters(type)

    def get_version(self) -> str:
        """
        Get provider config version

        :return: provider config version
        """
        return self.get_provider().get_version()

    def log(self, msg: str):
        """
        Log info message

        :param msg: message
        """
        is_log = False
        if self.window.core.config.has("log.llama") \
                and self.window.core.config.get("log.llama"):
            is_log = True
        self.window.core.debug.info(msg, not is_log)
        if is_log:
            print("[LLAMA-INDEX] {}".format(msg))
        self.window.idx_logger_message.emit(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.15 01:00:00                  #
# ================================================== #

import os
import uuid
from typing import Optional, List, Dict, Tuple, Union

import requests

from bs4 import BeautifulSoup
from urllib.parse import urljoin


class Helpers:
    def __init__(self, window=None):
        """
        Web helpers core

        :param window: Window instance
        """
        self.window = window

    def request(
            self,
            url: str = "",
            method: str = "GET",
            headers: Optional[dict] = None,
            params: Optional[dict] = None,
            data: Optional[Union[str, dict]] = None,
            json: Optional[dict] = None,
            files: Optional[dict] = None,
            cookies: Optional[dict] = None,
            timeout: int = 10,
            disable_ssl_verify: bool = False,
            allow_redirects: bool = True,
            stream: bool = False,
            user_agent: Optional[str] = None,
    ) -> Tuple[Optional[int], Optional[str]]:
        """
        Make HTTP request

        :param url: URL
        :param method: HTTP method
        :param headers: Headers
        :param params: GET parameters
        :param data: POST data
        :param json: JSON data
        :param files: Files
        :param cookies: Cookies
        :param timeout: Timeout
        :param disable_ssl_verify: Disable SSL verification
        :param allow_redirects: Allow redirects
        :param stream: Stream
        :param user_agent: User agent
        :return: status code, response text
        """
        upload = {}
        try:
            method = method.upper()
            session = requests.Session()
            args = {}

            if data:
                args['data'] = data
            if json:
                args['json'] = json
            if cookies:
                args['cookies'] = cookies
            if params:
                args['params'] = params
            if headers:
                args['headers'] = headers

            args['timeout'] = timeout
            if disable_ssl_verify:
                args['verify'] = False
            if not allow_redirects:
                args['allow_redirects'] = False
            if stream:
                args['stream'] = True
            if user_agent:
                if 'headers' not in args:
                    args['headers'] = {}
                args['headers']['User-Agent'] = user_agent

            if files:
                for key, value in files.items():
                    if os.path.exists(value) and os.path.isfile(value):
                        upload[key] = open(value, 'rb')
                args['files'] = upload

            if method == 'GET':
                response = session.get(url, **args)
            elif method == 'POST':
                response = session.post(url, **args)
            elif method == 'PUT':
                response = session.put(url, **args)
            elif method == 'DELETE':
                response = session.delete(url, **args)
            elif method == 'PATCH':
                response = session.patch(url, **args)
            else:
                return None, f'Invalid HTTP method: {method}'
            for k in upload:
                upload[k].close()  # close files if opened
            return response.status_code, response.text
        except Exception as e:
            for k in upload:
                upload[k].close()  # close files if opened
            return None, f'Error: {e}'

    def get_main_image(self, url: str) -> Optional[str]:
        """
        Get main image from URL

        :param url: URL to get image from
        :return: image URL
        """
        response = requests.get(url)
        soup = BeautifulSoup(response.content, 'html.parser')

        og_image = soup.find('meta', property='og:image')
        if og_image and og_image.get('content'):
            return og_image['content']

        twitter_image = soup.find('meta', attrs={'name': 'twitter:image'})
        if twitter_image and twitter_image.get('content'):
            return twitter_image['content']

        link_image = soup.find('link', rel='image_src')
        if link_image and link_image.get('href'):
            return link_image['href']

        images = soup.find_all('img')
        if images:
            images = [img for img in images if 'logo' not in (img.get('src') or '').lower()]
            largest_image = None
            max_area = 0
            for img in images:
                src = img.get('src')
                if not src:
                    continue
                src = requests.compat.urljoin(url, src)
                try:
                    img_response = requests.get(src, stream=True, timeout=5)
                    img_response.raw.decode_content = True

                    from PIL import Image
                    image = Image.open(img_response.raw)
                    width, height = image.size
                    area = width * height
                    if area > max_area:
                        max_area = area
                        largest_image = src
                except:
                    continue
            if largest_image:
                return largest_image
        return None

    def get_links(self, url: str) -> List[Dict[str, str]]:
        """
        Get links from URL

        :param url: URL to get links from
        :return: links list
        """
        response = requests.get(url)
        soup = BeautifulSoup(response.content, 'html.parser')
        links = []
        urls = []
        for link in soup.find_all('a'):
            try:
                name = link.get_text(strip=True)
                address = link.get('href')
                if address:
                    address = urljoin(url, address)
                    if not name:
                        title = link.get('title')
                        if title:
                            name = title
                        else:
                            name = address
                    if address not in urls:
                        urls.append(address)
                        links.append({name: address})
            except:
                continue
        return links


    def get_images(self, url: str) -> List[str]:
        """
        Get images from URL

        :param url: URL to get images from
        :return: images list
        """
        response = requests.get(url)
        soup = BeautifulSoup(response.content, 'html.parser')
        images = []
        for img in soup.find_all('img'):
            try:
                address = img.get('src')
                if address:
                    address = urljoin(url, address)
                    if address not in images:
                        images.append(address)
            except:
                continue
        return images

    def download_image(self, img: str) -> str:
        """
        Download image from URL

        :param img: URL to download image from
        :return: local path to image
        """
        dir = self.window.core.config.get_user_dir("img")
        response = requests.get(img, stream=True)
        name = img.replace("http://", "").replace("https://", "").replace("/", "_")
        path = os.path.join(dir, name)
        if os.path.exists(path):
            name = name + uuid.uuid4().hex[:6].upper()
        download_path = os.path.join(dir, name)
        with open(download_path, 'wb', ) as f:
            f.write(response.content)
        return self.window.core.filesystem.make_local(download_path)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, Dict, Any, List

from llama_index.core.tools import FunctionTool

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.bridge.worker import BridgeSignals
from pygpt_net.core.events import Event, KernelEvent
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Runner:
    def __init__(self, window=None):
        """
        Agent runner

        :param window: Window instance
        """
        self.window = window
        self.signals = None  # BridgeSignals
        self.error = None  # last exception
        self.next_instruction = ""  # evaluation instruction
        self.prev_score = -1  # evaluation score

    def call(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            signals: BridgeSignals
    ) -> bool:
        """
        Call an agent

        :param context: BridgeContext
        :param extra: extra data
        :param signals: BridgeSignals
        :return: True if success
        """
        if self.is_stopped():
            return True  # abort if stopped

        id = extra.get("agent_provider", "openai")
        verbose = self.window.core.config.get("agent.llama.verbose", False)
        try:
            # prepare input context
            ctx = context.ctx
            ctx.extra["agent_input"] = True  # mark as user input
            ctx.agent_call = True  # disables reply from plugin commands
            prompt = context.prompt

            # prepare agent
            model = context.model
            max_steps = self.window.core.config.get("agent.llama.steps", 10)
            tools = self.window.core.agents.tools.prepare(context, extra)
            history = self.window.core.agents.memory.prepare(context)
            llm = self.window.core.idx.llm.get(model)

            provider = None
            agent = None
            if self.window.core.agents.provider.has(id):
                kwargs = {
                    "context": context,
                    "tools": tools,
                    "llm": llm,
                    "chat_history": history,
                    "max_iterations": max_steps,
                    "verbose": verbose,
                }
                provider = self.window.core.agents.provider.get(id)
                agent = provider.get_agent(self.window, kwargs)
                if verbose:
                    print("Using Agent: " + id + ", model: " + model.id)

            if agent is None:
                raise Exception("Agent not found: " + id)

            # run agent
            mode = provider.get_mode()
            kwargs = {
                "agent": agent,
                "ctx": ctx,
                "prompt": prompt,
                "signals": signals,
                "verbose": verbose,
            }
            if mode == "plan":
                return self.run_plan(**kwargs)
            elif mode == "step":
                return self.run_steps(**kwargs)
            elif mode == "assistant":
                return self.run_assistant(**kwargs)

        except Exception as err:
            self.window.core.debug.error(err)
            self.error = err
            return False

    def run_steps(
            self,
            agent: Any,
            ctx: CtxItem,
            prompt: str,
            signals: BridgeSignals,
            verbose: bool = False
    ) -> bool:
        """
        Run agent steps

        :param agent: Agent instance
        :param ctx: Input context
        :param prompt: input text
        :param signals: BridgeSignals
        :param verbose: verbose mode
        :return: True if success
        """
        if self.is_stopped():
            return True  # abort if stopped

        is_last = False
        task = agent.create_task(self.prepare_input(prompt))
        tools_output = None

        # run steps
        i = 1
        idx = 0
        while not is_last:
            if self.is_stopped():
                break  # handle force stop

            if verbose:
                print ("\n----------- BEGIN STEP {} ----------\n".format(i))

            self.set_busy(signals)
            step_output = agent.run_step(task.task_id)
            is_last = step_output.is_last

            # append each step to chat output, last step = final response, so we skip it
            tools_output = self.window.core.agents.tools.export_sources(step_output.output)

            # get only current step tool calls using idx
            if tools_output:
                # check if idx is in range
                if idx < len(tools_output):
                    tools_output = [tools_output[idx]]
                    # INFO: idx is indexed from 0
                    # because tool outputs from previous step goes to the next ctx item!

            if verbose:
                print("\n")
                print("Step: " + str(i))
                print("Is last: " + str(is_last))
                print("Tool calls: " + str(tools_output))
                print("\n")

            if not is_last:
                step_ctx = self.add_ctx(ctx)
                step_ctx.set_input(str(tools_output))
                step_ctx.set_output("`{step_label} {i}`".format(
                    step_label=trans('msg.agent.step'),
                    i=str(i)
                ))
                step_ctx.cmds = tools_output
                step_ctx.results = ctx.results  # get results from base ctx
                ctx.results = []  # reset results

                # copy extra data (output from plugins)
                if tools_output:
                    for k in ctx.extra:
                        if not k.startswith("agent_"):
                            step_ctx.extra[k] = ctx.extra[k]

                # reset input ctx
                for k in list(ctx.extra.keys()):
                    if k != "agent_input":
                        del ctx.extra[k]

                step_ctx.extra["agent_step"] = True
                self.send_response(step_ctx, signals, KernelEvent.APPEND_DATA)
            i += 1
            idx += 1

        # final response
        if is_last:
            if self.is_stopped():
                return True  # abort if stopped
            response = agent.finalize_response(task.task_id)
            response_ctx = self.add_ctx(ctx)
            response_ctx.set_input(str(tools_output))
            response_ctx.set_output(str(response))
            response_ctx.extra["agent_output"] = True  # mark as output response
            response_ctx.extra["agent_finish"] = True  # mark as finished
            self.send_response(response_ctx, signals, KernelEvent.APPEND_DATA)
        self.set_idle(signals)
        return True

    def run_assistant(
            self,
            agent: Any,
            ctx: CtxItem,
            prompt: str,
            signals: BridgeSignals,
            verbose: bool = False
    ) -> bool:
        """
        Run agent steps

        :param agent: Agent instance
        :param ctx: Input context
        :param prompt: input text
        :param signals: BridgeSignals
        :param verbose: verbose mode
        :return: True if success
        """
        if self.is_stopped():
            return True  # abort if stopped

        if verbose:
            print("Running OpenAI Assistant...")
            print("Assistant: " + str(agent.assistant.id))
            print("Thread: " + str(agent.thread_id))
            print("Model: " + str(agent.assistant.model))
            print("Input: " + str(ctx.input))

        thread_id = agent.thread_id
        ctx.meta.thread = thread_id
        ctx.meta.assistant = agent.assistant.id
        ctx.thread = thread_id

        if self.is_stopped():
            return True  # abort if stopped

        # final response
        self.set_busy(signals)
        response = agent.chat(self.prepare_input(prompt))
        response_ctx = self.add_ctx(ctx)
        response_ctx.thread = thread_id
        response_ctx.set_input("Assistant")
        response_ctx.set_output(str(response))
        response_ctx.extra["agent_output"] = True  # mark as output response
        response_ctx.extra["agent_finish"] = True  # mark as finished
        self.send_response(response_ctx, signals, KernelEvent.APPEND_DATA)
        return True

    def run_plan(
            self,
            agent: Any,
            ctx: CtxItem,
            prompt: str,
            signals: BridgeSignals,
            verbose: bool = False
    ) -> bool:
        """
        Run agent sub-tasks

        :param agent: Agent instance
        :param ctx: Input context
        :param prompt: input text
        :param signals: BridgeSignals
        :param verbose: verbose mode
        :return: True if success
        """
        if self.is_stopped():
            return True  # abort if stopped

        self.set_busy(signals)
        plan_id = agent.create_plan(
            self.prepare_input(prompt)
        )
        plan = agent.state.plan_dict[plan_id]
        c = len(plan.sub_tasks)

        # prepare plan description
        plan_desc = "`{current_plan}:`\n".format(
            current_plan=trans('msg.agent.plan.current')
        )
        i = 1
        for sub_task in plan.sub_tasks:
            plan_desc += "\n\n"
            plan_desc += "\n**===== {sub_task_label}: {sub_task_name} =====**".format(
                sub_task_label=trans('msg.agent.plan.subtask'),
                sub_task_name=sub_task.name,
            )
            plan_desc += "\n{expected_label}: {expected_output}".format(
                expected_label=trans('msg.agent.plan.expected'),
                expected_output=str(sub_task.expected_output),
            )
            plan_desc += "\n{deps_label}: {dependencies}".format(
                deps_label=trans('msg.agent.plan.deps'),
                dependencies=str(sub_task.dependencies),
            )
            i += 1

        if verbose:
            print(plan_desc)

        # -----------------------------------------------------------

        step_ctx = self.add_ctx(ctx)
        step_ctx.set_input("{num_subtasks_label}: {count}, {plan_label}: {plan_id}".format(
            num_subtasks_label=trans('msg.agent.plan.num_subtasks'),
            count=str(c),
            plan_label=trans('msg.agent.plan'),
            plan_id=plan_id,
        ))
        step_ctx.set_output(plan_desc)
        step_ctx.cmds = []
        self.send_response(step_ctx, signals, KernelEvent.APPEND_DATA)  # send plan description

        i = 1
        for sub_task in plan.sub_tasks:
            if self.is_stopped():
                break

            self.set_busy(signals)

            j = 1
            task = agent.state.get_task(sub_task.name)

            self.set_status(signals, trans("status.agent.reasoning"))
            step_output = agent.run_step(task.task_id)
            tools_output = self.window.core.agents.tools.export_sources(step_output.output)

            task_header = "\n"
            task_header += "\n**===== Sub Task {index}: {sub_task_name} =====**".format(
                index=str(i),
                sub_task_name=sub_task.name
            )
            task_header += "\nExpected output: {expected_output}".format(
                expected_output=str(sub_task.expected_output)
            )
            task_header += "\nDependencies: {dependencies}".format(
                dependencies=str(sub_task.dependencies)
            )

            if verbose:
                print(task_header)

            # this can be the last step in current sub-task, so send it first
            step_ctx = self.add_ctx(ctx)
            step_ctx.set_input(str(tools_output)) # TODO: tool_outputs?
            step_ctx.set_output("`{sub_task_label} {i}/{c}, {step_label} {j}`\n{task_header}".format(
                sub_task_label=trans('msg.agent.plan.subtask'),
                i=str(i),
                c=str(c),
                step_label=trans('msg.agent.step'),
                j=str(j),
                task_header=task_header
            ))
            self.copy_step_results(step_ctx, ctx, tools_output)
            step_ctx.extra["agent_step"] = True
            self.send_response(step_ctx, signals, KernelEvent.APPEND_DATA)

            # -----------------------------------------------------------

            # loop until the last step is reached, if first step is not last
            while not step_output.is_last:
                if self.is_stopped():
                    break

                self.set_busy(signals)

                step_output = agent.run_step(task.task_id)
                tools_output = self.window.core.agents.tools.export_sources(step_output.output)

                # do not send again the first step (it was sent before the loop)
                if j > 1:
                    step_ctx = self.add_ctx(ctx)
                    step_ctx.set_input(str(tools_output))
                    step_ctx.set_output("`{sub_task_label} {i}/{c}, {step_label} {j}`\n{task_header}".format(
                        sub_task_label=trans('msg.agent.plan.subtask'),
                        i=str(i),
                        c=str(c),
                        step_label=trans('msg.agent.step'),
                        j=str(j),
                        task_header=task_header
                    ))
                    self.copy_step_results(step_ctx, ctx, tools_output)
                    step_ctx.extra["agent_step"] = True
                    self.send_response(step_ctx, signals, KernelEvent.APPEND_DATA)
                j += 1

            # finalize the response and commit to memory
            extra = {}
            extra["agent_output"] = True,  # mark as output response (will be attached to ctx items on continue)

            if i == c:  # if last subtask
                extra["agent_finish"] = True

            if self.is_stopped():
                return True  # abort if stopped

            # the is no tool calls here, only final response
            if step_output.is_last:
                response = agent.finalize_response(task.task_id, step_output=step_output)
                response_ctx = self.add_ctx(ctx)
                response_ctx.set_input(str(tools_output))
                response_ctx.set_output(str(response))
                response_ctx.extra.update(extra)  # extend with `agent_output` and `agent_finish`
                self.send_response(response_ctx, signals, KernelEvent.APPEND_DATA)
            i += 1

        return True

    def copy_step_results(
            self,
            step_ctx: CtxItem,
            ctx: CtxItem,
            tools_output: List[Dict[str, Any]]
    ):
        """
        Copy step results from base context item

        INFO: It is required to copy results from previous context item to the current one
        because tool calls store their output in the base context item, so we need to copy it

        :param step_ctx: previous context
        :param ctx: current context
        :param tools_output: tools output
        """
        step_ctx.cmds = tools_output
        step_ctx.results = ctx.results

        # copy extra data (output from plugins)
        if tools_output:
            for k in ctx.extra:
                if not k.startswith("agent_"):  # do not copy `agent_input` and etc.
                    step_ctx.extra[k] = ctx.extra[k]  # copy only plugin output

        # reset input ctx, remove outputs from plugins (tools) from previous step
        for k in list(ctx.extra.keys()):
            if k != "agent_input":
                del ctx.extra[k]

    def run_once(
            self,
            input: str,
            tools: List[FunctionTool],
            model_name: Optional[str] = None
    ) -> str:
        """
        Run agent once (quick call to ReAct agent)

        :param input: input text
        :param tools: tools
        :param model_name: model name
        :return: text response
        """
        if self.is_stopped():
            return ""  # abort if stopped

        verbose = self.window.core.config.get("agent.llama.verbose", False)
        model = self.window.core.models.get(model_name)
        llm = self.window.core.idx.llm.get(model)
        kwargs = {
            "context": BridgeContext(),
            "tools": tools,
            "llm": llm,
            "chat_history": [],
            "max_iterations": 10,
            "verbose": verbose,
        }
        provider = self.window.core.agents.provider.get("react")
        agent = provider.get_agent(self.window, kwargs)
        return agent.chat(self.prepare_input(input))

    def run_next(
            self,
            context: BridgeContext,
            extra: dict,
            signals: BridgeSignals
    ) -> bool:
        """
        Evaluate a response and run next step

        :param context: BridgeContext
        :param extra: extra data
        :param signals: BridgeSignals
        """
        if self.is_stopped():
            return True  # abort if stopped

        ctx = context.ctx
        self.send_response(ctx, signals, KernelEvent.APPEND_BEGIN)  # lock input, show stop btn
        history = context.history
        prompt = self.window.core.agents.observer.evaluation.get_prompt(history)
        tools = self.window.core.agents.observer.evaluation.get_tools()

        # evaluate
        self.set_busy(signals)
        self.next_instruction = ""  # reset
        self.prev_score = -1  # reset
        self.run_once(prompt, tools, ctx.model)  # tool will update evaluation
        return self.handle_evaluation(ctx, self.next_instruction, self.prev_score, signals)

    def handle_evaluation(
            self,
            ctx: CtxItem,
            instruction: str,
            score: int,
            signals: BridgeSignals
    ):
        """
        Handle evaluation

        :param ctx: CtxItem
        :param instruction: instruction
        :param score: score
        :param signals: BridgeSignals
        """
        if self.is_stopped():
            return True  # abort if stopped

        score = int(score)
        msg = "{score_label}: {score}%".format(
            score_label=trans('eval.score'),
            score=str(score)
        )
        self.set_status(signals, msg)
        if score < 0:
            self.send_response(ctx, signals, KernelEvent.APPEND_END)
            return True
        good_score = self.window.core.config.get("agent.llama.loop.score", 75)
        if score >= good_score != 0:
            msg = "{status_finished} {score_label}: {score}%".format(
                status_finished=trans('status.finished'),
                score_label=trans('eval.score'),
                score=str(score)
            )
            self.send_response(ctx, signals, KernelEvent.APPEND_END, msg=msg)
            return True

        # print("Instruction: " + instruction, "Score: " + str(score))
        step_ctx = self.add_ctx(ctx)
        step_ctx.set_input(instruction)
        step_ctx.set_output("")
        step_ctx.results = [
            {
                "loop": {
                    "score": score,
                }
            }
        ]
        step_ctx.extra = {
            "agent_input": True,
            "agent_evaluate": True,
            "footer": "Score: " + str(score) + "%",
        }
        step_ctx.internal = False  # input

        self.set_busy(signals)
        self.send_response(step_ctx, signals, KernelEvent.APPEND_DATA)

        # call next run
        context = BridgeContext()
        context.ctx = step_ctx
        context.history = self.window.core.ctx.all(meta_id=ctx.meta.id)
        preset = self.window.controller.presets.get_current()
        extra = {
            "agent_idx": preset.idx,
            "agent_provider": preset.agent_provider,
        }
        context.model = self.window.core.models.get(self.window.core.config.get('model'))
        return self.call(context, extra, signals)

    def add_ctx(
            self,
            from_ctx: CtxItem
    ) -> CtxItem:
        """
        Add context item

        :param from_ctx: CtxItem (parent, source)
        :return: CtxItem
        """
        ctx = CtxItem()
        ctx.meta = from_ctx.meta
        ctx.internal = True
        ctx.current = True  # mark as current context item
        ctx.mode = from_ctx.mode
        ctx.model = from_ctx.model
        ctx.prev_ctx = from_ctx
        ctx.images = from_ctx.images  # copy from parent if appended from plugins
        ctx.urls = from_ctx.urls  # copy from parent if appended from plugins
        ctx.attachments = from_ctx.attachments # copy from parent if appended from plugins
        ctx.live = True
        return ctx

    def send_response(
            self,
            ctx: CtxItem,
            signals: BridgeSignals,
            event_name: str,
            **kwargs
    ):
        """
        Send async response to chat window (BridgeSignals)

        :param ctx: CtxItem
        :param signals: BridgeSignals
        :param event_name: kernel event
        :param kwargs: extra data
        """
        context = BridgeContext()
        context.ctx = ctx
        event = KernelEvent(event_name, {
            'context': context,
            'extra': kwargs,
        })
        signals.response.emit(event)

    def set_busy(
            self,
            signals: BridgeSignals,
            **kwargs
    ):
        """
        Set busy status

        :param signals: BridgeSignals
        :param kwargs: extra data
        """
        data = {
            "id": "agent",
            "msg": trans("status.agent.reasoning"),
        }
        event = KernelEvent(KernelEvent.STATE_BUSY, data)
        data.update(kwargs)
        signals.response.emit(event)

    def set_idle(
            self,
            signals: BridgeSignals,
            **kwargs
    ):
        """
        Set idle status

        :param signals: BridgeSignals
        :param kwargs: extra data
        """
        data = {
            "id": "agent",
        }
        event = KernelEvent(KernelEvent.STATE_IDLE, data)
        data.update(kwargs)
        signals.response.emit(event)

    def set_status(
            self,
            signals: BridgeSignals,
            msg: str
    ):
        """
        Set busy status

        :param signals: BridgeSignals
        :param msg: status message
        """
        data = {
            "status": msg,
        }
        event = KernelEvent(KernelEvent.STATUS, data)
        signals.response.emit(event)

    def prepare_input(self, prompt: str) -> str:
        """
        Prepare input context

        :param prompt: input text
        """
        event = Event(Event.AGENT_PROMPT, {
            'value': prompt,
        })
        self.window.dispatch(event)
        return event.data['value']

    def is_stopped(self) -> bool:
        """
        Check if run is stopped

        :return: True if stopped
        """
        return self.window.controller.kernel.stopped()

    def get_error(self) -> Optional[Exception]:
        """
        Get last error

        :return: last exception or None if no error
        """
        return self.error



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import List

from pygpt_net.core.types import (
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_AUDIO,
    MODE_RESEARCH,
)

class Legacy:
    def __init__(self, window=None):
        """
        Agents core (legacy)

        :param window: Window instance
        """
        self.window = window
        self.allowed_modes = [
            MODE_CHAT,
            MODE_COMPLETION,
            MODE_VISION,
            MODE_LANGCHAIN,
            MODE_LLAMA_INDEX,
            MODE_AUDIO,
            MODE_RESEARCH,
        ]

    def get_allowed_modes(self) -> List[str]:
        """
        Get allowed modes

        :return: allowed modes
        """
        return self.allowed_modes

    def get_mode(self) -> str:
        """
        Get sub-mode to use internally

        :return: sub-mode
        """
        mode = MODE_CHAT
        current = self.window.core.config.get("agent.mode")
        if current is not None and current in self.allowed_modes:
            mode = current
        return mode

    def get_idx(self) -> str:
        """
        Get agent index

        :return: agent index
        """
        return self.window.core.config.get("agent.idx")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.30 04:00:00                  #
# ================================================== #

import copy

from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItemModel, QAction, QIcon
from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QScrollArea, QWidget, QTabWidget, QFrame, \
    QSplitter, QSizePolicy, QMenuBar, QCheckBox

from pygpt_net.ui.widget.dialog.assistant_store import AssistantVectorStoreDialog
from pygpt_net.ui.widget.element.group import CollapsedGroup
from pygpt_net.ui.widget.element.labels import UrlLabel
from pygpt_net.ui.widget.lists.assistant_store import AssistantVectorStoreEditorList
from pygpt_net.ui.widget.option.checkbox import OptionCheckbox
from pygpt_net.ui.widget.option.combo import OptionCombo
from pygpt_net.ui.widget.option.dictionary import OptionDict
from pygpt_net.ui.widget.option.input import OptionInput, PasswordInput
from pygpt_net.ui.widget.option.slider import OptionSlider
from pygpt_net.ui.widget.option.textarea import OptionTextarea
from pygpt_net.utils import trans


class AssistantVectorStore:
    def __init__(self, window=None):
        """
        Assistant vector store editor dialog

        :param window: Window instance
        """
        self.window = window
        self.dialog_id = "assistant.store"

    def setup(self, idx=None):
        """
        Setup vector store editor dialog

        :param idx: current model tab index
        """
        self.window.ui.nodes['assistant.store.btn.new'] = \
            QPushButton(trans("dialog.assistant.store.btn.new"))
        self.window.ui.nodes['assistant.store.btn.save'] = \
            QPushButton(trans("dialog.assistant.store.btn.save"))
        self.window.ui.nodes['assistant.store.btn.refresh_status'] = \
            QPushButton(QIcon(":/icons/reload.svg"), "")
        self.window.ui.nodes['assistant.store.btn.close'] = \
            QPushButton(trans("dialog.assistant.store.btn.close"))

        self.window.ui.nodes['assistant.store.btn.refresh_status'].setToolTip(
            trans("dialog.assistant.store.btn.refresh_status")
        )
        self.window.ui.nodes['assistant.store.btn.refresh_status'].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        self.window.ui.nodes['assistant.store.btn.upload.files'] = \
            QPushButton(trans("dialog.assistant.store.btn.upload.files"))
        self.window.ui.nodes['assistant.store.btn.upload.dir'] = \
            QPushButton(trans("dialog.assistant.store.btn.upload.dir"))

        self.window.ui.nodes['assistant.store.hide_thread'] = QCheckBox(trans("assistant.store.hide_threads"))
        self.window.ui.nodes['assistant.store.hide_thread'].setChecked(True)
        self.window.ui.nodes['assistant.store.hide_thread'].stateChanged.connect(
            lambda: self.window.controller.assistant.store.set_hide_thread(
                self.window.ui.nodes['assistant.store.hide_thread'].isChecked()
            )
        )
        if not self.window.core.config.get("assistant.store.hide_threads"):
            self.window.ui.nodes['assistant.store.hide_thread'].setChecked(False)

        self.window.ui.nodes['assistant.store.btn.new'].clicked.connect(
            lambda: self.window.controller.assistant.store.new()
        )
        self.window.ui.nodes['assistant.store.btn.save'].clicked.connect(
            lambda: self.window.controller.assistant.store.save()
        )
        self.window.ui.nodes['assistant.store.btn.close'].clicked.connect(
            lambda: self.window.controller.assistant.store.close()
        )
        self.window.ui.nodes['assistant.store.btn.refresh_status'].clicked.connect(
            lambda: self.window.controller.assistant.store.refresh_status()
        )
        self.window.ui.nodes['assistant.store.btn.upload.files'].clicked.connect(
            lambda: self.window.controller.assistant.batch.open_upload_files()
        )
        self.window.ui.nodes['assistant.store.btn.upload.dir'].clicked.connect(
            lambda: self.window.controller.assistant.batch.open_upload_dir()
        )

        # set enter key to save button
        self.window.ui.nodes['assistant.store.btn.new'].setAutoDefault(False)
        self.window.ui.nodes['assistant.store.btn.refresh_status'].setAutoDefault(False)
        self.window.ui.nodes['assistant.store.btn.save'].setAutoDefault(True)

        # footer buttons
        footer = QHBoxLayout()
        footer.addWidget(self.window.ui.nodes['assistant.store.btn.close'])
        footer.addWidget(self.window.ui.nodes['assistant.store.btn.save'])

        # upload buttons
        upload_layout = QHBoxLayout()
        upload_layout.addWidget(self.window.ui.nodes['assistant.store.btn.refresh_status'])  # reload status btn
        upload_layout.addWidget(self.window.ui.nodes['assistant.store.btn.upload.files'])
        upload_layout.addWidget(self.window.ui.nodes['assistant.store.btn.upload.dir'])
        upload_layout.setContentsMargins(0, 0, 0, 0)

        # editor tabs
        self.window.ui.tabs['assistant.store'] = QTabWidget()

        # build settings tabs
        parent_id = "assistant.store"

        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        content = QVBoxLayout()

        # create config options entry if not exists
        if parent_id not in self.window.ui.config:
            self.window.ui.config[parent_id] = {}

        # get options
        options = copy.deepcopy(self.window.controller.assistant.store.get_options())
        widgets = self.build_widgets(options)
        advanced_keys = []
        for key in options:
            if 'advanced' in options[key] and options[key]['advanced']:
                advanced_keys.append(key)

        # apply settings widgets
        for key in widgets:
            self.window.ui.config[parent_id][key] = widgets[key]

        for key in widgets:
            if key in advanced_keys:  # hide advanced options
                continue
            content.addLayout(self.add_option(widgets[key], options[key]))  # add to scroll

        # append advanced options at the end
        if len(advanced_keys) > 0:
            group_id = 'assistant.store.advanced'
            self.window.ui.groups[group_id] = CollapsedGroup(self.window, group_id, None, False, None)
            self.window.ui.groups[group_id].box.setText(trans('settings.advanced.collapse'))
            for key in widgets:
                if key not in advanced_keys:  # ignore non-advanced options
                    continue

                option = self.add_option(widgets[key], options[key])  # build option
                self.window.ui.groups[group_id].add_layout(option)  # add option to group

            # add advanced options group to scroll
            content.addWidget(self.window.ui.groups[group_id])


        content.addLayout(upload_layout)  # upload buttons
        content.setContentsMargins(0, 0, 0, 0)

        # scroll widget
        scroll_widget = QWidget()
        scroll_widget.setLayout(content)
        scroll.setWidget(scroll_widget)

        area = QVBoxLayout()
        area.addWidget(scroll)
        area.setContentsMargins(0, 0, 0, 0)

        area_widget = QWidget()
        area_widget.setLayout(area)

        data = {}
        for store_id in self.window.core.assistants.store.items.keys():
            store = self.window.core.assistants.store.items[store_id]
            data[store_id] = store

        # storage list
        id = 'assistant.store.list'
        self.window.ui.nodes[id] = AssistantVectorStoreEditorList(self.window, id)
        self.window.ui.models[id] = self.create_model(self.window)
        self.window.ui.nodes[id].setModel(self.window.ui.models[id])

        # left widget
        left_layout = QVBoxLayout()
        left_layout.addWidget(self.window.ui.nodes['assistant.store.btn.new'])
        left_layout.addWidget(self.window.ui.nodes[id])
        left_layout.addWidget(self.window.ui.nodes['assistant.store.hide_thread'])

        left_layout.setContentsMargins(0, 0, 0, 0)
        left_widget = QWidget()
        left_widget.setLayout(left_layout)

        # update models list
        self.update_list(id, data)

        # set max width to list
        self.window.ui.nodes[id].setMinimumWidth(250)

        # splitter
        self.window.ui.splitters['dialog.assistant.store'] = QSplitter(Qt.Horizontal)
        self.window.ui.splitters['dialog.assistant.store'].addWidget(left_widget)  # list
        self.window.ui.splitters['dialog.assistant.store'].addWidget(area_widget)  # tabs
        self.window.ui.splitters['dialog.assistant.store'].setStretchFactor(0, 2)
        self.window.ui.splitters['dialog.assistant.store'].setStretchFactor(1, 5)
        self.window.ui.splitters['dialog.assistant.store'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        main_layout = QHBoxLayout()
        main_layout.addWidget(self.window.ui.splitters['dialog.assistant.store'])
        main_layout.setContentsMargins(0, 0, 0, 0)

        layout = QVBoxLayout()
        layout.addLayout(main_layout)  # list + model tabs
        layout.addLayout(footer)  # bottom buttons (save, defaults)

        # add menu
        layout.setMenuBar(self.setup_menu())

        self.window.ui.dialog[self.dialog_id] = AssistantVectorStoreDialog(self.window, self.dialog_id)
        self.window.ui.dialog[self.dialog_id].setLayout(layout)
        self.window.ui.dialog[self.dialog_id].setWindowTitle(trans('dialog.assistant.store'))

        # restore current opened tab if idx is set
        if idx is not None:
            try:
                self.window.controller.assistant.store.set_by_tab(idx)
            except Exception as e:
                print('Failed restore store editor tab: {}'.format(idx))
        else:
            if self.window.controller.assistant.store.current is None:
                self.window.controller.assistant.store.set_by_tab(0)

    def setup_menu(self) -> QMenuBar:
        """
        Setup menu bar

        :return: QMenuBar
        """
        self.actions = {}
        self.menu_bar = QMenuBar()

        self.menu = {}
        self.menu["current"] = self.menu_bar.addMenu(trans("dialog.assistant.store.menu.current"))
        self.menu["all"] = self.menu_bar.addMenu(trans("dialog.assistant.store.menu.all"))

        # --------------------------------------------

        # import files (current)
        self.actions["current.import_files"] = QAction(QIcon(":/icons/download.svg"),
                                                   trans("dialog.assistant.store.menu.current.import_files"))
        self.actions["current.import_files"].triggered.connect(
            lambda: self.window.controller.assistant.batch.import_store_files(
                self.window.controller.assistant.store.current  # current selected store
            )
        )

        # refresh (current)
        self.actions["current.refresh_store"] = QAction(QIcon(":/icons/reload.svg"),
                                                     trans("dialog.assistant.store.menu.current.refresh_store"))
        self.actions["current.refresh_store"].triggered.connect(
            lambda: self.window.controller.assistant.store.refresh_status()
        )

        # clear files (current, local)
        self.actions["current.clear_files"] = QAction(QIcon(":/icons/close.svg"),
                                                  trans("dialog.assistant.store.menu.current.clear_files"))
        self.actions["current.clear_files"].triggered.connect(
            lambda: self.window.controller.assistant.batch.clear_store_files(
                self.window.controller.assistant.store.current  # current selected store
            )
        )

        # truncate files (current, local + remote)
        self.actions["current.truncate_files"] = QAction(QIcon(":/icons/delete.svg"),
                                                     trans("dialog.assistant.store.menu.current.truncate_files"))
        self.actions["current.truncate_files"].triggered.connect(
            lambda: self.window.controller.assistant.batch.truncate_store_files(
                self.window.controller.assistant.store.current  # current selected store
            )
        )

        # delete (current, local + remote)
        self.actions["current.delete"] = QAction(QIcon(":/icons/delete.svg"),
                                                         trans("dialog.assistant.store.menu.current.delete"))
        self.actions["current.delete"].triggered.connect(
            lambda: self.window.controller.assistant.store.delete(
                self.window.controller.assistant.store.current  # current selected store
            )
        )

        # --------------------------------------------

        # import all (stores + files)
        self.actions["all.import_all"] = QAction(QIcon(":/icons/download.svg"), trans("dialog.assistant.store.menu.all.import_all"))
        self.actions["all.import_all"].triggered.connect(
            lambda: self.window.controller.assistant.batch.import_stores()
        )

        # import files (all)
        self.actions["all.import_files"] = QAction(QIcon(":/icons/download.svg"),
                                             trans("dialog.assistant.store.menu.all.import_files"))
        self.actions["all.import_files"].triggered.connect(
            lambda: self.window.controller.assistant.batch.import_files()
        )

        # refresh (local)
        self.actions["all.refresh_stores"] = QAction(QIcon(":/icons/reload.svg"),
                                               trans("dialog.assistant.store.menu.all.refresh_store"))
        self.actions["all.refresh_stores"].triggered.connect(
            lambda: self.window.controller.assistant.batch.refresh_stores()
        )

        # clear stores (all, local)
        self.actions["all.clear_stores"] = QAction(QIcon(":/icons/close.svg"),
                                                  trans("dialog.assistant.store.menu.all.clear_store"))
        self.actions["all.clear_stores"].triggered.connect(
            lambda: self.window.controller.assistant.batch.clear_stores()
        )

        # clear files (all, local)
        self.actions["all.clear_files"] = QAction(QIcon(":/icons/close.svg"),
                                               trans("dialog.assistant.store.menu.all.clear_files"))
        self.actions["all.clear_files"].triggered.connect(
            lambda: self.window.controller.assistant.batch.clear_files()
        )

        # truncate stores (all, local + remote)
        self.actions["all.truncate_stores"] = QAction(QIcon(":/icons/delete.svg"), trans("dialog.assistant.store.menu.all.truncate_store"))
        self.actions["all.truncate_stores"].triggered.connect(
            lambda: self.window.controller.assistant.batch.truncate_stores()
        )

        # truncate files (all, local + remote)
        self.actions["all.truncate_files"] = QAction(QIcon(":/icons/delete.svg"), trans("dialog.assistant.store.menu.all.truncate_files"))
        self.actions["all.truncate_files"].triggered.connect(
            lambda: self.window.controller.assistant.batch.truncate_files()
        )

        # current
        self.menu["current"].addAction(self.actions["current.import_files"])
        self.menu["current"].addAction(self.actions["current.refresh_store"])
        self.menu["current"].addAction(self.actions["current.clear_files"])
        self.menu["current"].addAction(self.actions["current.truncate_files"])
        self.menu["current"].addAction(self.actions["current.delete"])

        # all
        self.menu["all"].addAction(self.actions["all.import_all"])
        self.menu["all"].addAction(self.actions["all.import_files"])
        self.menu["all"].addAction(self.actions["all.refresh_stores"])
        self.menu["all"].addAction(self.actions["all.clear_stores"])
        self.menu["all"].addAction(self.actions["all.clear_files"])
        self.menu["all"].addAction(self.actions["all.truncate_stores"])
        self.menu["all"].addAction(self.actions["all.truncate_files"])

        return self.menu_bar

    def build_widgets(self, options: dict) -> dict:
        """
        Build settings options widgets

        :param options: model options
        :return: dict of widgets
        """
        parent = "model"
        widgets = {}

        for key in options:
            option = options[key]
            option['id'] = key
            # create widget by option type
            if option['type'] == 'text' or option['type'] == 'int' or option['type'] == 'float':
                if 'slider' in option and option['slider'] \
                        and (option['type'] == 'int' or option['type'] == 'float'):
                    widgets[key] = OptionSlider(self.window, parent, key, option)  # slider + text input
                else:
                    if 'secret' in option and option['secret']:
                        widgets[key] = PasswordInput(self.window, parent, key, option)  # password input
                    else:
                        widgets[key] = OptionInput(self.window, parent, key, option)  # text input
            elif option['type'] == 'textarea':
                widgets[key] = OptionTextarea(self.window, parent, key, option)  # textarea
            elif option['type'] == 'bool':
                widgets[key] = OptionCheckbox(self.window, parent, key, option)  # checkbox
            elif option['type'] == 'dict':
                self.window.controller.config.placeholder.apply(option)
                widgets[key] = OptionDict(self.window, parent, key, option)  # dictionary
                widgets[key].setMinimumHeight(200)
            elif option['type'] == 'combo':
                self.window.controller.config.placeholder.apply(option)
                widgets[key] = OptionCombo(self.window, parent, key, option)  # combobox

        return widgets

    def add_line(self) -> QFrame:
        """
        Make separator line

        :return: separator line
        """
        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        return line

    def add_urls(self, urls: dict) -> QWidget:
        """
        Add clickable urls to list

        :param urls: urls dict
        :return: QWidget
        """
        layout = QVBoxLayout()
        for name in urls:
            url = urls[name]
            label = UrlLabel(name, url)
            layout.addWidget(label)

        widget = QWidget()
        widget.setLayout(layout)
        return widget

    def add_option(self, widget: QWidget, option: dict) -> QVBoxLayout:
        """
        Append option widget to layout

        :param widget: widget instance
        :param option: option dict
        :return: QVBoxLayout
        """
        one_column_types = ['textarea', 'dict', 'bool']
        key = option['id']
        label = trans(option['label'])
        label_key = 'model.' + key + '.label'
        desc = None
        desc_key = None

        if option['type'] != 'bool':
            self.window.ui.nodes[label_key] = QLabel(label)
            self.window.ui.nodes[label_key].setStyleSheet("font-weight: bold;")

        if "description" in option:
            desc = trans(option['description'])
            desc_key = 'model.' + key + '.desc'

        # 2-columns layout
        if option['type'] not in one_column_types:
            cols = QHBoxLayout()
            cols.addWidget(self.window.ui.nodes[label_key])  # disable label in bool type
            cols.addWidget(widget)

            cols_widget = QWidget()
            cols_widget.setLayout(cols)
            cols_widget.setMaximumHeight(90)

            layout = QVBoxLayout()
            layout.addWidget(cols_widget)
        else:
            # 1-column layout: textarea and dict fields
            layout = QVBoxLayout()
            if option['type'] != 'bool':
                layout.addWidget(self.window.ui.nodes[label_key])
            else:
                widget.box.setText(label)  # set checkbox label
            layout.addWidget(widget)

        if desc:
            self.window.ui.nodes[desc_key] = QLabel(desc)
            self.window.ui.nodes[desc_key].setWordWrap(True)
            self.window.ui.nodes[desc_key].setMaximumHeight(40)
            self.window.ui.nodes[desc_key].setStyleSheet("font-size: 10px;")
            layout.addWidget(self.window.ui.nodes[desc_key])

        line = self.add_line()  # TODO: change name to separator
        layout.addWidget(line)

        return layout

    def add_raw_option(self, option: dict) -> QHBoxLayout:
        """
        Add raw option row

        :param option: Option
        :return: QHBoxLayout
        """
        layout = QHBoxLayout()
        layout.addWidget(option)
        return layout

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model
        :param parent: parent widget
        :return: QStandardItemModel
        """
        return QStandardItemModel(0, 1, parent)

    def update_list(self, id: str, data: dict):
        """
        Update list

        :param id: ID of the list
        :param data: Data to update
        """
        if id not in self.window.ui.models:
            return
        self.window.ui.models[id].removeRows(0, self.window.ui.models[id].rowCount())
        i = 0
        for n in data:
            name = data[n].name
            extra = ""
            if name == "" or name is None:
                if self.window.core.config.get("assistant.store.hide_threads"):
                    continue
                name = data[n].id
            if data[n].num_files > 0:
                extra += "{} {}".format(data[n].num_files, trans("assistant.store.files.suffix"))
            if data[n].usage_bytes > 0:
                extra += " - {}".format(self.window.core.filesystem.sizeof_fmt(data[n].usage_bytes))
            if extra != "":
                name += " ({})".format(extra)
            self.window.ui.models[id].insertRow(i)
            self.window.ui.models[id].setData(self.window.ui.models[id].index(i, 0), name)
            i += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.26 19:00:00                  #
# ================================================== #

import re


def sanitize_name(name: str) -> str:
    """
    Sanitize name

    :param name: name
    :return: sanitized name
    """
    if name is None:
        return ""
    # allowed characters: a-z, A-Z, 0-9, _, and -
    name = name.strip().lower()
    sanitized_name = re.sub(r'[^a-z0-9_-]', '_', name)
    return sanitized_name[:64]  # limit to 64 characters


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Tuple, Optional, Dict, List

from PySide6.QtCore import Qt, QSize
from PySide6.QtGui import QColor


class Common:
    def __init__(self, window=None):
        """
        Painter common methods controller

        :param window: Window instance
        """
        self.window = window

    def convert_to_size(self, canvas_size: str) -> Tuple[int, int]:
        """
        Convert string to size

        :param canvas_size: Canvas size string
        :return: tuple (width, height)
        """
        return tuple(map(int, canvas_size.split('x'))) if canvas_size else (0, 0)

    def set_canvas_size(self, width: int, height: int):
        """
        Set canvas size

        :param width: int
        :param height: int
        """
        self.window.ui.painter.setFixedSize(QSize(width, height))

    def set_brush_mode(self, enabled: bool):
        """
        Set the paint mode

        :param enabled: bool
        """
        if enabled:
            self.window.ui.nodes['painter.select.brush.color'].setCurrentText("Black")
            self.window.ui.painter.set_brush_color(Qt.black)
            self.window.core.config.set('painter.brush.mode', "brush")
            self.window.core.config.save()

    def set_erase_mode(self, enabled: bool):
        """
        Set the erase mode

        :param enabled: bool
        """
        if enabled:
            self.window.ui.nodes['painter.select.brush.color'].setCurrentText("White")
            self.window.ui.painter.set_brush_color(Qt.white)
            self.window.core.config.set('painter.brush.mode', "erase")
            self.window.core.config.save()

    def change_canvas_size(self, selected: Optional[str] = None):
        """
        Change the canvas size

        :param selected: Selected size
        """
        if not selected:
            selected = self.window.ui.nodes['painter.select.canvas.size'].currentData()
        if selected:
            size = self.convert_to_size(selected)
            self.window.ui.nodes['painter.select.canvas.size'].setCurrentText(selected)
            self.set_canvas_size(size[0], size[1])
            self.window.core.config.set('painter.canvas.size', selected)
            self.window.core.config.save()

    def change_brush_size(self, size: int):
        """
        Change the brush size

        :param size: Brush size
        """
        size = int(size)
        self.update_brush_size(size)
        self.window.core.config.set('painter.brush.size', size)
        self.window.core.config.save()

    def change_brush_color(self):
        """Change the brush color"""
        color = self.window.ui.nodes['painter.select.brush.color'].currentData()
        text_color = self.window.ui.nodes['painter.select.brush.color'].currentText()
        self.update_brush_color(color)
        self.window.core.config.set('painter.brush.color', text_color)
        self.window.core.config.save()

    def update_brush_size(self, size: int):
        """
        Update brush size

        :param size: Brush size
        """
        self.window.ui.painter.set_brush_size(size)

    def update_brush_color(self, color: QColor):
        """
        Update brush color

        :param color: QColor
        """
        self.window.ui.painter.set_brush_color(color)

    def restore_brush_settings(self):
        """Restore brush settings"""
        brush_color = None
        if self.window.core.config.has('painter.brush.color'):
            brush_color = self.window.core.config.get('painter.brush.color', "Black")

        # mode
        if self.window.core.config.has('painter.brush.mode'):
            mode = self.window.core.config.get('painter.brush.mode', "brush")
            if mode == "brush":
                self.window.ui.nodes['painter.btn.brush'].setChecked(True)
                self.set_brush_mode(True)
            elif mode == "erase":
                self.window.ui.nodes['painter.btn.erase'].setChecked(True)
                self.set_erase_mode(True)
        # color
        if brush_color:
            color = self.get_colors().get(brush_color)
            self.window.ui.nodes['painter.select.brush.color'].setCurrentText(brush_color)
            self.update_brush_color(color)

        # size
        size = 3
        if self.window.core.config.has('painter.brush.size'):
            size = int(self.window.core.config.get('painter.brush.size', 3))
        self.window.ui.nodes['painter.select.brush.size'].setCurrentIndex(
                self.window.ui.nodes['painter.select.brush.size'].findText(str(size))
        )

    def get_colors(self) -> Dict[str, QColor]:
        """
        Get colors dict

        :return: colors dict
        """
        return {
            "Black": Qt.black,
            "White": Qt.white,
            "Red": Qt.red,
            "Orange": QColor('orange'),
            "Yellow": Qt.yellow,
            "Green": Qt.green,
            "Blue": Qt.blue,
            "Indigo": QColor('indigo'),
            "Violet": QColor('violet')
        }

    def get_sizes(self) -> List[str]:
        """
        Get brush sizes

        :return: list of sizes
        """
        return ['1', '2', '3', '5', '8', '12', '15', '20', '25', '30', '50', '100', '200']

    def get_canvas_sizes(self) -> List[str]:
        """
        Get canvas sizes

        :return: list of sizes
        """
        return [
            # horizontal
            "640x480", "800x600", "1024x768", "1280x720", "1600x900",
            "1920x1080", "2560x1440", "3840x2160", "4096x2160",
            # vertical
            "480x640", "600x800", "768x1024", "720x1280", "900x1600",
            "1080x1920", "1440x2560", "2160x3840", "2160x4096"
        ]

    def get_capture_dir(self) -> str:
        """
        Get capture directory

        :return: path to capture directory
        """
        return self.window.core.config.get_user_dir('capture')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import re

from PySide6.QtCore import QTimer

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_ASSISTANT,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION, 
    MODE_IMAGE,
)
from pygpt_net.core.tabs.tab import Tab
from pygpt_net.core.events import ControlEvent
from pygpt_net.utils import trans


class Control:

    CTX_TIMER_DELAY = 2000  # delay before ctx change execute (ms)
    MODE_TIMER_DELAY = 2000  # delay before mode change execute (ms)

    def __init__(self, window=None):
        """
        Control handler

        :param window: Window instance
        """
        self.window = window
        self.last_confirm = None
        self.ctx_timer = QTimer(self.window)
        self.ctx_timer.timeout.connect(self.handle_ctx)
        self.mode_timer = QTimer(self.window)
        self.mode_timer.timeout.connect(self.handle_mode)
        self.ctx_action = None
        self.mode_action = None

    def handle_ctx(self):
        """Handle context action (delayed)"""
        self.ctx_timer.stop()
        if self.ctx_action is not None:
            if self.ctx_action == ControlEvent.CTX_NEW:
                self.window.controller.ctx.new(force=True)
            elif self.ctx_action == ControlEvent.CTX_PREV:
                self.window.controller.ctx.prev()
            elif self.ctx_action == ControlEvent.CTX_NEXT:
                self.window.controller.ctx.next()
            elif self.ctx_action == ControlEvent.CTX_LAST:
                self.window.controller.ctx.last()
            self.ctx_action = None

    def handle_mode(self):
        """Handle mode action (delayed)"""
        self.mode_timer.stop()
        if self.mode_action is not None:
            if self.mode_action == ControlEvent.MODE_NEXT:
                self.window.controller.mode.next()
            elif self.mode_action == ControlEvent.MODE_PREV:
                self.window.controller.mode.prev()
            elif self.mode_action == ControlEvent.MODE_CHAT:
                self.window.controller.mode.set(MODE_CHAT)
            elif self.mode_action == ControlEvent.MODE_LLAMA_INDEX:
                self.window.controller.mode.set(MODE_LLAMA_INDEX)
            self.mode_action = None

    def handle(self, event: ControlEvent, force: bool = False):
        """
        Handle accessibility event (ControlEvent)

        :param event: event object
        :param force: True if force window close on APP_EXIT event
        """
        self.window.core.debug.info("EVENT CTRL: " + event.name)

        # reset last confirm
        if event.name != ControlEvent.CMD_CONFIRM:
            self.last_confirm = None

        # app exit
        if event.name == ControlEvent.APP_EXIT:
            if force:  # if confirmed
                self.window.close()
            else:  # play confirm
                self.last_confirm = event
                self.window.controller.audio.play_event(
                    trans("event.audio.confirm"),
                    event,
                )
            return
        # status changed
        elif event.name == ControlEvent.APP_STATUS:
            status = self.window.core.access.helpers.get_status()
            if status != "":
                self.window.controller.audio.play_event(
                    status,
                    event,
                )

        # unrecognized command (abort)
        if event.name == ControlEvent.VOICE_CONTROL_UNRECOGNIZED:
            self.window.controller.audio.play_event(
                trans("event.audio.voice.control.unrecognized"),
                event,
            )
            return

        # confirm last action
        elif event.name == ControlEvent.CMD_CONFIRM:
            if self.last_confirm is not None:
                self.handle(self.last_confirm, force=True)
                self.last_confirm = None

        # camera
        elif event.name == ControlEvent.CAMERA_ENABLE:
            self.window.controller.camera.enable_capture()
        elif event.name == ControlEvent.CAMERA_DISABLE:
            self.window.controller.camera.disable_capture()
        elif event.name == ControlEvent.CAMERA_CAPTURE:
            self.window.controller.camera.manual_capture()

        # ctx
        elif event.name == ControlEvent.CTX_NEW:
            self.ctx_action = ControlEvent.CTX_NEW
            self.ctx_timer.start(self.CTX_TIMER_DELAY)
        elif event.name == ControlEvent.CTX_PREV:
            self.ctx_action = ControlEvent.CTX_PREV
            self.ctx_timer.start(self.CTX_TIMER_DELAY)
        elif event.name == ControlEvent.CTX_NEXT:
            self.ctx_action = ControlEvent.CTX_NEXT
            self.ctx_timer.start(self.CTX_TIMER_DELAY)
        elif event.name == ControlEvent.CTX_LAST:
            self.ctx_action = ControlEvent.CTX_LAST
            self.ctx_timer.start(self.CTX_TIMER_DELAY)
        elif event.name == ControlEvent.CTX_INPUT_FOCUS:
            self.window.controller.chat.common.focus_input()
            self.handle_result(event, True)
        elif event.name == ControlEvent.CTX_INPUT_CLEAR:
            self.window.controller.chat.input.clear_input()
            self.handle_result(event, True)
        elif event.name == ControlEvent.CTX_STOP:
            self.window.controller.kernel.stop()
        elif event.name == ControlEvent.CTX_ATTACHMENTS_CLEAR:
            self.window.controller.attachment.clear(force=True)
        elif event.name == ControlEvent.CTX_CURRENT:
            status = self.window.core.access.helpers.get_current_ctx()
            if status != "":
                self.window.controller.audio.play_event(
                    status,
                    event,
                )
        elif event.name == ControlEvent.CTX_READ_LAST:
            text = self.window.core.access.helpers.get_last_ctx_item()
            if text != "":
                self.window.controller.audio.play_event(
                    text,
                    event,
                )
        elif event.name == ControlEvent.CTX_READ_ALL:
            text = self.window.core.access.helpers.get_all_ctx_items()
            if text != "":
                self.window.controller.audio.play_event(
                    text,
                    event,
                )
        elif event.name == ControlEvent.CTX_RENAME:
            if event.data is not None and "params" in event.data:
                msg = event.data["params"]
                if msg != "":
                    self.window.controller.ctx.update_name_current(msg)
                    msg = trans("event.audio.ctx.rename").format(ctx=msg)
                    self.window.controller.audio.play_event(
                        msg,
                        event,
                    )
        elif event.name == ControlEvent.CTX_SEARCH_STRING:
            if event.data is not None and "params" in event.data:
                msg = event.data["params"]
                if msg != "":
                    self.window.controller.ctx.append_search_string(msg)
                    num = self.window.core.ctx.count_found_meta()
                    msg = trans("event.audio.ctx.search.string").format(num=num)
                    self.window.controller.audio.play_event(
                        msg,
                        event,
                    )
        elif event.name == ControlEvent.CTX_SEARCH_CLEAR:
                self.window.controller.ctx.search_string_clear()
                self.handle_result(event, True)

        # calendar
        elif event.name == ControlEvent.CALENDAR_ADD:
            if event.data is not None and "params" in event.data:
                msg = event.data["params"]
                if msg != "":
                    self.window.controller.calendar.note.append_text(msg)
                    self.handle_result(event, True)
        elif event.name == ControlEvent.CALENDAR_CLEAR:
            if force:
                self.window.controller.calendar.note.clear_note()
                self.handle_result(event, True)
            else:
                self.last_confirm = event
                self.window.controller.audio.play_event(
                    trans("event.audio.confirm"),
                    event,
                )
        elif event.name == ControlEvent.CALENDAR_READ:
            text = self.window.controller.calendar.note.get_note_text()
            self.window.controller.audio.play_event(
                text,
                event,
            )

        # mode change
        elif event.name == ControlEvent.MODE_CHAT:
            self.mode_action = ControlEvent.MODE_CHAT
            self.mode_timer.start(self.MODE_TIMER_DELAY)
        elif event.name == ControlEvent.MODE_LLAMA_INDEX:
            self.mode_action = ControlEvent.MODE_LLAMA_INDEX
            self.mode_timer.start(self.MODE_TIMER_DELAY)
        elif event.name == ControlEvent.MODE_NEXT:
            self.mode_action = ControlEvent.MODE_NEXT
            self.mode_timer.start(self.MODE_TIMER_DELAY)
        elif event.name == ControlEvent.MODE_PREV:
            self.mode_action = ControlEvent.MODE_PREV
            self.mode_timer.start(self.MODE_TIMER_DELAY)

        # model change
        elif event.name == ControlEvent.MODEL_NEXT:
            self.window.controller.model.next()
        elif event.name == ControlEvent.MODEL_PREV:
            self.window.controller.model.prev()

        # preset change
        elif event.name == ControlEvent.PRESET_NEXT:
            self.window.controller.presets.next()
        elif event.name == ControlEvent.PRESET_PREV:
            self.window.controller.presets.prev()

        # tab change
        elif event.name == ControlEvent.TAB_NEXT:
            self.window.controller.ui.tabs.next_tab()
        elif event.name == ControlEvent.TAB_PREV:
            self.window.controller.ui.tabs.prev_tab()

        # tabs: by names
        elif event.name == ControlEvent.TAB_CHAT:
            self.window.controller.ui.tabs.switch_tab(Tab.TAB_CHAT)
        elif event.name == ControlEvent.TAB_FILES:
            self.window.controller.ui.tabs.switch_tab(Tab.TAB_FILES)
        elif event.name == ControlEvent.TAB_CALENDAR:
            self.window.controller.ui.tabs.switch_tab(Tab.TAB_TOOL_CALENDAR)
        elif event.name == ControlEvent.TAB_DRAW:
            self.window.controller.ui.tabs.switch_tab(Tab.TAB_TOOL_PAINTER)

        # tabs: notepads
        elif event.name == ControlEvent.TAB_NOTEPAD:
            tmp_idx = 1
            if event.data is not None and "params" in event.data and event.data["params"] != "":
                try:
                    # regex extract number
                    num = re.findall(r'\d+', event.data["params"])
                    if len(num) > 0:
                        tmp_idx = int(num[0])
                except Exception as e:
                    print(e)
            self.window.controller.notepad.switch_to_tab(tmp_idx)
            self.handle_result(event, True)

        # voice control
        elif event.name == ControlEvent.VOICE_COMMAND_START:
            if not self.window.core.config.get("access.voice_control"):
                self.window.core.config.set("access.voice_control", True)
                self.window.controller.access.voice.update()
            self.window.controller.access.voice.start_recording()
        elif event.name == ControlEvent.VOICE_COMMAND_STOP:
            self.window.controller.access.voice.stop_recording()
        elif event.name == ControlEvent.VOICE_COMMAND_TOGGLE:
            if not self.window.core.config.get("access.voice_control"):
                self.window.core.config.set("access.voice_control", True)
                self.window.controller.access.voice.update()
            self.window.controller.access.voice.toggle_recording()

        # voice input
        elif event.name == ControlEvent.VOICE_MESSAGE_START:
            self.window.core.plugins.get("audio_input").handler_simple.start_recording()
        elif event.name == ControlEvent.VOICE_MESSAGE_STOP:
            self.window.core.plugins.get("audio_input").handler_simple.stop_recording()
        elif event.name == ControlEvent.VOICE_MESSAGE_TOGGLE:
            self.window.core.plugins.get("audio_input").handler_simple.toggle_recording()

        # audio enable/disable
        elif event.name == ControlEvent.AUDIO_INPUT_ENABLE:
            self.window.controller.plugins.enable('audio_input')
            self.handle_result(event, True)
        elif event.name == ControlEvent.AUDIO_INPUT_DISABLE:
            self.window.controller.plugins.disable('audio_input')
            self.handle_result(event, True)
        elif event.name == ControlEvent.AUDIO_OUTPUT_ENABLE:
            self.window.controller.plugins.enable('audio_output')
            self.handle_result(event, True)
        elif event.name == ControlEvent.AUDIO_OUTPUT_DISABLE:
            self.window.controller.plugins.disable('audio_output')
            self.handle_result(event, True)

        # text input
        elif event.name == ControlEvent.INPUT_SEND:
            if event.data is not None and "params" in event.data:
                msg = event.data["params"]
                if msg != "":
                    self.window.controller.chat.common.clear_input()
                    self.window.controller.chat.common.append_to_input(msg)
                self.window.controller.chat.input.send_input()
        elif event.name == ControlEvent.INPUT_APPEND:
            if event.data is not None and "params" in event.data:
                msg = event.data["params"]
                if msg != "":
                    self.window.controller.chat.common.append_to_input(msg)
                    self.handle_result(event, True)

        # notepad
        elif event.name == ControlEvent.NOTE_ADD:
            if event.data is not None and "params" in event.data:
                msg = event.data["params"]
                if msg != "":
                    if self.window.controller.notepad.is_active():
                        idx = self.window.controller.notepad.get_current_active()
                        self.window.controller.notepad.append_text(msg, idx)
                    else:
                        self.window.controller.notepad.append_text(msg, 1)
                    self.handle_result(event, True)
        elif event.name == ControlEvent.NOTEPAD_CLEAR:
            if force:
                if event.data is not None and "params" in event.data and event.data["params"] != "":
                    idx = 1
                    try:
                        # regex extract number
                        num = re.findall(r'\d+', event.data["params"])
                        if len(num) > 0:
                            idx = int(num[0])
                    except:
                        pass
                    if self.window.controller.notepad.clear(idx):
                        self.handle_result(event, True)
                else:
                    idx = self.window.controller.notepad.get_current_active()
                    if idx is not None and self.window.controller.notepad.clear(idx):
                        self.handle_result(event, True)
            else:
                self.last_confirm = event
                self.window.controller.audio.play_event(
                    trans("event.audio.confirm"),
                    event,
                )
        elif event.name == ControlEvent.NOTEPAD_READ:
            if event.data is not None and "params" in event.data and event.data["params"] != "":
                idx = 1
                try:
                    # regex extract number
                    num = re.findall(r'\d+', event.data["params"])
                    if len(num) > 0:
                        idx = int(num[0])
                except:
                    pass
                text = self.window.controller.notepad.get_notepad_text(idx)
            else:
                text = self.window.controller.notepad.get_current_notepad_text()
            self.window.controller.audio.play_event(
                text,
                event,
            )

        # cmd list (help)
        elif event.name == ControlEvent.CMD_LIST:
            text = self.window.core.access.voice.get_commands_string(values=True)
            self.window.controller.audio.play_event(
                text,
                event,
            )

    def handle_result(self, event, result: bool = True):
        """
        Play result

        :param event: event object
        :param result: result
        """
        if not self.window.core.config.get("access.audio.event.speech"):
            return
        if self.window.core.access.voice.is_muted(event.name):
            return
        self.window.controller.audio.play_event(
            trans("event.audio." + event.name),
            event,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.26 23:00:00                  #
# ================================================== #

import json
import time
from typing import Optional, Dict, Any, List

from pygpt_net.core.types import (
    MODE_CHAT,
    MODE_VISION,
    MODE_AUDIO,
    MODE_RESEARCH,
)
from pygpt_net.core.bridge.context import BridgeContext, MultimodalContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem

from .utils import sanitize_name
from pygpt_net.item.attachment import AttachmentItem


class Chat:
    def __init__(self, window=None):
        """
        Chat wrapper

        :param window: Window instance
        """
        self.window = window
        self.input_tokens = 0
        self.audio_prev_id = None
        self.audio_prev_expires_ts = None

    def send(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ):
        """
        Call OpenAI API for chat

        :param context: Bridge context
        :param extra: Extra arguments
        :return: response or stream chunks
        """
        prompt = context.prompt
        stream = context.stream
        max_tokens = int(context.max_tokens or 0)
        system_prompt = context.system_prompt
        mode = context.mode
        model = context.model
        functions = context.external_functions
        attachments = context.attachments
        multimodal_ctx = context.multimodal_ctx

        ctx = context.ctx
        if ctx is None:
            ctx = CtxItem()  # create empty context
        user_name = ctx.input_name  # from ctx
        ai_name = ctx.output_name  # from ctx

        client = self.window.core.gpt.get_client(mode)

        # build chat messages
        messages = self.build(
            prompt=prompt,
            system_prompt=system_prompt,
            model=model,
            history=context.history,
            attachments=attachments,
            ai_name=ai_name,
            user_name=user_name,
            multimodal_ctx=multimodal_ctx,
        )
        msg_tokens = self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        # check if max tokens not exceeded
        if max_tokens > 0:
            if msg_tokens + int(max_tokens) > model.ctx:
                max_tokens = model.ctx - msg_tokens - 1
                if max_tokens < 0:
                    max_tokens = 0

        # extra API kwargs
        response_kwargs = {}

        # tools / functions
        tools = []
        if functions is not None and isinstance(functions, list):
            for function in functions:
                if str(function['name']).strip() == '' or function['name'] is None:
                    continue
                params = {}
                if function['params'] is not None and function['params'] != "":
                    params = json.loads(function['params'])  # unpack JSON from string
                tools.append({
                    "type": "function",
                    "function": {
                        "name": function['name'],
                        "parameters": params,
                        "description": function['desc'],
                    }
                })

        # fix: o1 compatibility
        if model.id is not None and not model.id.startswith("o1") and not model.id.startswith("o3"):
            response_kwargs['presence_penalty'] = self.window.core.config.get('presence_penalty')
            response_kwargs['frequency_penalty'] = self.window.core.config.get('frequency_penalty')
            response_kwargs['temperature'] = self.window.core.config.get('temperature')
            response_kwargs['top_p'] = self.window.core.config.get('top_p')

        # extra arguments, o3 only
        if model.extra and "reasoning_effort" in model.extra:
            response_kwargs['reasoning_effort'] = model.extra["reasoning_effort"]

        # tool calls are not supported for o1-mini and o1-preview
        if (model.id is not None
                and model.id not in ["o1-mini", "o1-preview"]):
            if len(tools) > 0:
                response_kwargs['tools'] = tools

        if max_tokens > 0:
            if not model.id.startswith("o1") and not model.id.startswith("o3"):
                response_kwargs['max_tokens'] = max_tokens
            else:
                response_kwargs['max_completion_tokens'] = max_tokens

        # audio mode
        if mode in [MODE_AUDIO]:
            stream = False
            voice_id = "alloy"
            tmp_voice = self.window.core.plugins.get_option("audio_output", "openai_voice")
            if tmp_voice:
                voice_id = tmp_voice
            response_kwargs["modalities"] = ["text", "audio"]
            response_kwargs["audio"] = {
                "voice": voice_id,
                "format": "wav"
            }

        response = client.chat.completions.create(
            messages=messages,
            model=model.id,
            stream=stream,
            **response_kwargs,
        )
        return response

    def build(
            self,
            prompt: str,
            system_prompt: str,
            model: ModelItem,
            history: Optional[List[CtxItem]] = None,
            attachments: Optional[Dict[str, AttachmentItem]] = None,
            ai_name: Optional[str] = None,
            user_name: Optional[str] = None,
            multimodal_ctx: Optional[MultimodalContext] = None,
    ) -> list:
        """
        Build list of chat messages

        :param prompt: user prompt
        :param system_prompt: system prompt
        :param history: history
        :param model: model item
        :param attachments: attachments
        :param ai_name: AI name
        :param user_name: username
        :param multimodal_ctx: Multimodal context
        :return: messages list
        """
        messages = []

        # tokens config
        mode = MODE_CHAT
        allowed_system = True
        if (model.id is not None
                and model.id in ["o1-mini", "o1-preview"]):
            allowed_system = False

        used_tokens = self.window.core.tokens.from_user(
            prompt,
            system_prompt,
        )  # threshold and extra included
        max_ctx_tokens = self.window.core.config.get('max_total_tokens')  # max context window

        # fit to max model tokens
        if max_ctx_tokens > model.ctx:
            max_ctx_tokens = model.ctx

        # input tokens: reset
        self.reset_tokens()

        # names fallback
        if ai_name is None or ai_name == "":
            ai_name = "assistant"
        if user_name is None or user_name == "":
            user_name = "user"

        # append system prompt
        if allowed_system:
            if system_prompt is not None and system_prompt != "":
                messages.append({"role": "system", "content": system_prompt})

        # append messages from context (memory)
        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model.id,
                mode,
                used_tokens,
                max_ctx_tokens,
            )
            for item in items:
                # input
                if item.final_input is not None and item.final_input != "":
                    messages.append({
                        "role": "user",
                        "name": sanitize_name(user_name),
                        "content": item.final_input,
                    })

                # output
                if item.final_output is not None and item.final_output != "":
                    msg = {
                        "role": "assistant",
                        "name": sanitize_name(ai_name),
                        "content": item.final_output,
                    }
                    # append previous audio ID
                    if MODE_AUDIO in model.mode:
                        if item.audio_id:
                            # at first check expires_at - expired audio throws error in API
                            current_timestamp = time.time()
                            audio_timestamp = int(item.audio_expires_ts) if item.audio_expires_ts else 0
                            if audio_timestamp and audio_timestamp > current_timestamp:
                                msg["audio"] = {
                                    "id": item.audio_id
                                }
                        elif self.audio_prev_id:
                            current_timestamp = time.time()
                            audio_timestamp = int(self.audio_prev_expires_ts) if self.audio_prev_expires_ts else 0
                            if audio_timestamp and audio_timestamp > current_timestamp:
                                msg["audio"] = {
                                    "id": self.audio_prev_id
                                }
                    messages.append(msg)

        # use vision and audio if available in current model
        content = str(prompt)
        if MODE_VISION in model.mode:
            content = self.window.core.gpt.vision.build_content(
                content=content,
                attachments=attachments,
            )
        if MODE_AUDIO in model.mode:
            content = self.window.core.gpt.audio.build_content(
                content=content,
                multimodal_ctx=multimodal_ctx,
            )

        # append current prompt
        messages.append({
            "role": "user",
            "content": content,
        })

        # input tokens: update
        self.input_tokens += self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        return messages

    def reset_tokens(self):
        """Reset input tokens counter"""
        self.input_tokens = 0

    def get_used_tokens(self) -> int:
        """
        Get input tokens counter

        :return: input tokens
        """
        return self.input_tokens
